@
@ GCat's Master System Emulator
@ Copyright (c) 2012 James Perry
@
@ This program is free software: you can redistribute it and/or modify
@ it under the terms of the GNU General Public License as published by
@ the Free Software Foundation, either version 3 of the License, or
@ (at your option) any later version.
@
@ This program is distributed in the hope that it will be useful,
@ but WITHOUT ANY WARRANTY; without even the implied warranty of
@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
@ GNU General Public License for more details.
@
@ You should have received a copy of the GNU General Public License
@ along with this program.  If not, see <http://www.gnu.org/licenses/>.
@

@
@ Main part of ARM Z80 emulation core
@

@============================================================
@
@ Macros to simplify instruction implementations
@
@============================================================
@ General
#define END_INSN(cycles) \
	adds	r10, r10, cycles ; \
	bpl	done_execute	 ; \
	b	next_instruction ; \

@ END_INSN_IX is only different because it stores r8 back in IX or IY
@ An instruction that reads IX/IY but doesn't write it can use normal
@ END_INSN for speed.
#define END_INSN_IX(cycles) \
	ldrb	r2, [r0, #74]	 ; \
	cmp	r2, #0		 ; \
	streqh	r8, [r0, #24]	 ; \
	strneh	r8, [r0, #26]	 ; \
	adds	r10, r10, cycles ; \
	bpl	done_execute	 ; \
	b	next_instruction ; \

@ Memory access
#define READ_BYTE(dest, addr, scratch1, scratch2) \
	mov	scratch1, addr, lsr #13 ; \
	ldr	scratch2,[r11, scratch1, lsl #2] ; \
	cmp	addr, #0x400		    ; \
	ldrlt	scratch2,[r11, #64]	    ; \
	mov	scratch1, #0xff00	    ; \
	orr	scratch1, scratch1, #0xfc   ; \
	cmp	addr, scratch1		    ; \
	ldrge	scratch2,[r11, #68]	    ; \
	bic	scratch1, addr, #0xe000     ; \
	ldrb	dest, [scratch2, scratch1]	     ; \

#define READ_SIGNED_BYTE(dest, addr, scratch1, scratch2) \
	mov	scratch1, addr, lsr #13 ; \
	ldr	scratch2,[r11, scratch1, lsl #2] ; \
	cmp	addr, #0x400		    ; \
	ldrlt	scratch2,[r11, #64]	    ; \
	mov	scratch1, #0xff00	    ; \
	orr	scratch1, scratch1, #0xfc   ; \
	cmp	addr, scratch1		    ; \
	ldrge	scratch2,[r11, #68]	    ; \
	bic	scratch1, addr, #0xe000     ; \
	ldrsb	dest, [scratch2, scratch1]	     ; \

#define WRITE_BYTE(src, addr, scratch1, scratch2) \
	mov	scratch1, addr, lsr #13 ; \
	add	scratch2, r11, #32	; \
	ldr	scratch2,[scratch2, scratch1, lsl #2] ; \
	mov	scratch1, #0xff00	    ; \
	orr	scratch1, scratch1, #0xfc   ; \
	cmp	addr, scratch1		    ; \
	ldrge	scratch2,[r11, #68]	    ; \
	bic	scratch1, addr, #0xe000 ; \
	strb	src, [scratch2, scratch1 ] ; \
	blge	update_paging		   ; \
	
	
@ 8-bit arithmetic and logical
#define INC8_HI(reg, scratch1) \
	add	reg, reg, #256	; \
	bic	reg, reg, #65536 ; \
	/* need to set flags 3, 5, S, Z, H, V and clear N */\
	/* first clear 3, 5, S, N, Z */\
	bic	r4, r4, #0xfe00 ; \
	/* set 3, 5, S according to result */\
	and	scratch1, reg, #0xa800 ; \
	orr	r4, r4, scratch1       ; \
	/* set Z */\
	tst	reg, #0xff00 	; \
	orreq	r4, r4, #0x4000 ; \
	/* set V */\
	and	scratch1, reg, #0xff00 ; \
	cmp	scratch1, #0x8000      ; \
	orreq	r4, r4, #0x400	       ; \
	/* set H */\
	tst	scratch1, #0x0f00	    ; \
	orreq	r4, r4, #0x1000		    ; \

	
#define INC8_LO(reg, scratch1) \
	add	scratch1, reg, #1	; \
	and	scratch1, scratch1, #0xff ; \
	bic	reg, reg, #0xff		  ; \
	orr	reg, reg, scratch1	  ; \
	/* need to set flags 3, 5, S, Z, H, V and clear N */\
	/* first clear 3, 5, S, N, Z */\
	bic	r4, r4, #0xfe00 ; \
	/* set 3, 5, S according to result */\
	and	scratch1, reg, #0xa8 ; \
	orr	r4, r4, scratch1, lsl #8   ; \
	/* set Z */\
	tst	reg, #0xff 	; \
	orreq	r4, r4, #0x4000 ; \
	/* set V */\
	and	scratch1, reg, #0xff ; \
	cmp	scratch1, #0x80    ; \
	orreq	r4, r4, #0x400	     ; \
	/* set H */\
	tst	scratch1, #0x0f	; \
	orreq	r4, r4, #0x1000	; \

	
#define DEC8_HI(reg, scratch1) \
	sub	reg, reg, #256	; \
	bic	reg, reg, #0xff000000 ; \
	bic	reg, reg, #0xff0000   ; \
	/* need to set flags 3, 5, S, Z, H, V and set N */\
	/* first clear 3, 5, S, N, Z */\
	bic	r4, r4, #0xfe00 ; \
	orr	r4, r4, #0x0200 ; \
	/* set 3, 5, S according to result */\
	and	scratch1, reg, #0xa800 ; \
	orr	r4, r4, scratch1       ; \
	/* set Z */\
	tst	reg, #0xff00 	; \
	orreq	r4, r4, #0x4000 ; \
	/* set V */\
	and	scratch1, reg, #0xff00 ; \
	cmp	scratch1, #0x7f00      ; \
	orreq	r4, r4, #0x400	       ; \
	/* set H */\
	and	scratch1, scratch1, #0x0f00 ; \
	cmp	scratch1, #0x0f00 ; \
	orreq	r4, r4, #0x1000		    ; \


#define DEC8_LO(reg, scratch1) \
	sub	scratch1, reg, #1	; \
	and	scratch1, scratch1, #0xff ; \
	bic	reg, reg, #0xff		  ; \
	orr	reg, reg, scratch1	  ; \
	/* need to set flags 3, 5, S, Z, H, V and clear N */\
	/* first clear 3, 5, S, N, Z */\
	bic	r4, r4, #0xfe00 ; \
	orr	r4, r4, #0x200 	; \
	/* set 3, 5, S according to result */\
	and	scratch1, reg, #0xa8 ; \
	orr	r4, r4, scratch1, lsl #8   ; \
	/* set Z */\
	tst	reg, #0xff 	; \
	orreq	r4, r4, #0x4000 ; \
	/* set V */\
	and	scratch1, reg, #0xff ; \
	cmp	scratch1, #0x7f	     ; \
	orreq	r4, r4, #0x400	     ; \
	/* set H */\
	and	scratch1, scratch1, #0x0f ; \
	cmp	scratch1, #0x0f		  ; \
	orreq	r4, r4, #0x1000		  ; \
	

#define RRC8_HI(reg, scratch1, scratch2) \
	mov	scratch1, reg, lsr #8 ; \
	mov	scratch2, scratch1, lsr #1 ; \
	orr	scratch2, scratch2, scratch1, lsl #7 ; \
	and	scratch2, scratch2, #0xff	     ; \
	/* do flags */\
	bic	r4, r4, #0xff00 ; \
	/* carry */\
	and	scratch1, scratch2, #0x80 ; \
	orr	r4, r4, scratch1, lsl #1  ; \
	/* other flags */\
	ldr	scratch1, [r0, #44] ; \
	ldrb	scratch1, [scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8       ; \
	/* store result back in register*/\
	bic	reg, reg, #0xff00 ; \
	orr	reg, reg, scratch2, lsl #8 ; \


#define RRC8_LO(reg, scratch1, scratch2) \
	and	scratch1, reg, #0xff ; \
	mov	scratch2, scratch1, lsr #1 ; \
	orr	scratch2, scratch2, scratch1, lsl #7 ; \
	and	scratch2, scratch2, #0xff	     ; \
	/* do flags */\
	bic	r4, r4, #0xff00 ; \
	/* carry */\
	and	scratch1, scratch2, #0x80 ; \
	orr	r4, r4, scratch1, lsl #1  ; \
	/* other flags */\
	ldr	scratch1, [r0, #44] ; \
	ldrb	scratch1, [scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8       ; \
	/* store result back in register*/\
	bic	reg, reg, #0xff ; \
	orr	reg, reg, scratch2 ; \

	
	
	@ RLC:
	@ - Carry set by bit 7 of source
	@ - H and N set to 0
	@ - S, Z, P, undoc set from logic table
#define RLC8_HI(reg, scratch1, scratch2) \
	mov	scratch1, reg, lsr #8 ; \
	mov	scratch2, scratch1, lsl #1 ; \
	orr	scratch2, scratch2, scratch1, lsr #7 ; \
	/* do flags */\
	bic	r4, r4, #0xff00 ; \
	/* carry */\
	and	scratch1, scratch2, #0x100 ; \
	and	scratch2, scratch2, #0xff  ; \
	orr	r4, r4, scratch1	   ; \
	/* other flags */\
	ldr	scratch1,[r0, #44] ; \
	ldrb	scratch1,[scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8      ; \
	/* store result back in register*/\
	bic	reg, reg, #0xff00		     ; \
	orr	reg, reg, scratch2, lsl #8	     ; \
	
	
#define RLC8_LO(reg, scratch1, scratch2) \
	and	scratch1, reg, #0xff ; \
	mov	scratch2, scratch1, lsl #1 ; \
	orr	scratch2, scratch2, scratch1, lsr #7 ; \
	/* do flags */\
	bic	r4, r4, #0xff00 ; \
	/* carry */\
	and	scratch1, scratch2, #0x100 ; \
	and	scratch2, scratch2, #0xff  ; \
	orr	r4, r4, scratch1	   ; \
	/* other flags */\
	ldr	scratch1,[r0, #44] ; \
	ldrb	scratch1,[scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8      ; \
	/* store result back in register*/\
	bic	reg, reg, #0xff		     ; \
	orr	reg, reg, scratch2	     ; \


#define RL8_HI(reg, scratch1, scratch2) \
	mov	scratch2, reg, lsr #7 ; \
	and	scratch2, scratch2, #0xfe ; \
	and	scratch1, r4, #0x100	  ; \
	orr	scratch2, scratch2, scratch1, lsr #8 ; \
	mov	scratch1, reg, lsr #7		     ; \
	and	scratch1, scratch1, #0x100	     ; \
	/* do flags */\
	bic	r4, r4, #0xff00 ; \
	/* carry */\
	orr	r4, r4, scratch1	   ; \
	/* other flags */\
	ldr	scratch1,[r0, #44] ; \
	ldrb	scratch1,[scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8      ; \
	/* store result back in register*/\
	bic	reg, reg, #0xff00		     ; \
	orr	reg, reg, scratch2, lsl #8	     ; \
	

#define RL8_LO(reg, scratch1, scratch2) \
	mov	scratch2, reg, lsl #1 ; \
	and	scratch2, scratch2, #0xfe ; \
	and	scratch1, r4, #0x100	  ; \
	orr	scratch2, scratch2, scratch1, lsr #8 ; \
	mov	scratch1, reg, lsl #1		     ; \
	and	scratch1, scratch1, #0x100	     ; \
	/* do flags */\
	bic	r4, r4, #0xff00 ; \
	/* carry */\
	orr	r4, r4, scratch1	   ; \
	/* other flags */\
	ldr	scratch1,[r0, #44] ; \
	ldrb	scratch1,[scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8      ; \
	/* store result back in register*/\
	bic	reg, reg, #0xff		     ; \
	orr	reg, reg, scratch2	     ; \

	
#define RR8_HI(reg, scratch1, scratch2) \
	mov 	scratch2, reg, lsr #9 ; \
	and	scratch1, r4, #0x100  ; \
	orr	scratch2, scratch2, scratch1, lsr #1 ; \
	and	scratch1, reg, #0x100		     ; \
	/* do flags */\
	bic	r4, r4, #0xff00 ; \
	/* carry */\
	orr	r4, r4, scratch1	   ; \
	/* other flags */\
	ldr	scratch1,[r0, #44] ; \
	ldrb	scratch1,[scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8      ; \
	/* store result back in register*/\
	bic	reg, reg, #0xff00		     ; \
	orr	reg, reg, scratch2, lsl #8	     ; \
	
		
#define RR8_LO(reg, scratch1, scratch2) \
	mov 	scratch2, reg, lsr #1 ; \
	and	scratch2, scratch2, #0x7f	     ; \
	and	scratch1, r4, #0x100  ; \
	orr	scratch2, scratch2, scratch1, lsr #1 ; \
	and	scratch1, reg, #0x1		     ; \
	/* do flags */\
	bic	r4, r4, #0xff00 ; \
	/* carry */\
	orr	r4, r4, scratch1, lsl #8	   ; \
	/* other flags */\
	ldr	scratch1,[r0, #44] ; \
	ldrb	scratch1,[scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8      ; \
	/* store result back in register*/\
	bic	reg, reg, #0xff		     ; \
	orr	reg, reg, scratch2	     ; \

	
#define SLA8_HI(reg, scratch1, scratch2) \
	mov	scratch2, reg, lsr #7 ; \
	bic	scratch2, scratch2, #1 ; \
	/* do flags */\
	bic	r4, r4, #0xff00 ; \
	/* carry */\
	and	scratch1, scratch2, #0x100 ; \
	and	scratch2, scratch2, #0xff  ; \
	orr	r4, r4, scratch1	   ; \
	/* other flags */\
	ldr	scratch1,[r0, #44] ; \
	ldrb	scratch1,[scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8      ; \
	/* store result back in register */\
	bic	reg, reg, #0xff00 ; \
	orr	reg, reg, scratch2, lsl #8 ; \
	
	
#define SLA8_LO(reg, scratch1, scratch2) \
	and	scratch2, reg, #0xff	      ; \
	mov	scratch2, scratch2, lsl #1    ; \
	/* do flags */\
	bic	r4, r4, #0xff00 ; \
	/* carry */\
	and	scratch1, scratch2, #0x100 ; \
	and	scratch2, scratch2, #0xff  ; \
	orr	r4, r4, scratch1	   ; \
	/* other flags */\
	ldr	scratch1,[r0, #44] ; \
	ldrb	scratch1,[scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8      ; \
	/* store result back in register */\
	bic	reg, reg, #0xff ; \
	orr	reg, reg, scratch2 ; \

	
#define SRA8_HI(reg, scratch1, scratch2) \
	mov	scratch2, reg, lsr #9 ; \
	mov	scratch1, reg, lsr #8 ; \
	and	scratch1, scratch1, #0x80      ; \
	orr	scratch2, scratch2, scratch1   ; \
	/* do flags */\
	bic	r4, r4, #0xff00 ; \
	/* carry */\
	and	scratch1, reg, #0x100 ; \
	orr	r4, r4, scratch1      ; \
	/* other flags */\
	ldr	scratch1, [r0, #44] ; \
	ldrb	scratch1, [scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8       ; \
	/* store result back in register */\
	bic	reg, reg, #0xff00 ; \
	orr	reg, reg, scratch2, lsl #8 ; \
	
	
#define SRA8_LO(reg, scratch1, scratch2) \
	and	scratch2, reg, #0xff ; \
	mov	scratch2, scratch2, lsr #1 ; \
	and	scratch1, reg, #0x80	   ; \
	orr	scratch2, scratch2, scratch1   ; \
	/* do flags */\
	bic	r4, r4, #0xff00 ; \
	/* carry */\
	and	scratch1, reg, #0x1 ; \
	orr	r4, r4, scratch1, lsl #8      ; \
	/* other flags */\
	ldr	scratch1, [r0, #44] ; \
	ldrb	scratch1, [scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8       ; \
	/* store result back in register */\
	bic	reg, reg, #0xff ; \
	orr	reg, reg, scratch2 ; \

	
#define SLL8_HI(reg, scratch1, scratch2) \
	mov	scratch2, reg, lsr #7 ; \
	orr	scratch2, scratch2, #1 ; \
	/* do flags */\
	bic	r4, r4, #0xff00 ; \
	/* carry */\
	and	scratch1, scratch2, #0x100 ; \
	and	scratch2, scratch2, #0xff  ; \
	orr	r4, r4, scratch1	   ; \
	/* other flags */\
	ldr	scratch1,[r0, #44] ; \
	ldrb	scratch1,[scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8      ; \
	/* store result back in register */\
	bic	reg, reg, #0xff00 ; \
	orr	reg, reg, scratch2, lsl #8 ; \

	
#define SLL8_LO(reg, scratch1, scratch2) \
	and	scratch2, reg, #0xff	      ; \
	mov	scratch2, scratch2, lsl #1    ; \
	orr	scratch2, scratch2, #1	      ; \
	/* do flags */\
	bic	r4, r4, #0xff00 ; \
	/* carry */\
	and	scratch1, scratch2, #0x100 ; \
	and	scratch2, scratch2, #0xff  ; \
	orr	r4, r4, scratch1	   ; \
	/* other flags */\
	ldr	scratch1,[r0, #44] ; \
	ldrb	scratch1,[scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8      ; \
	/* store result back in register */\
	bic	reg, reg, #0xff ; \
	orr	reg, reg, scratch2 ; \

	
#define SRL8_HI(reg, scratch1, scratch2) \
	mov	scratch2, reg, lsr #9 ; \
	/* do flags */\
	bic	r4, r4, #0xff00 ; \
	/* carry */\
	and	scratch1, reg, #0x100 ; \
	orr	r4, r4, scratch1      ; \
	/* other flags */\
	ldr	scratch1, [r0, #44] ; \
	ldrb	scratch1, [scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8       ; \
	/* store result back in register */\
	bic	reg, reg, #0xff00 ; \
	orr	reg, reg, scratch2, lsl #8 ; \

	
#define SRL8_LO(reg, scratch1, scratch2) \
	and	scratch2, reg, #0xff ; \
	mov	scratch2, scratch2, lsr #1 ; \
	/* do flags */\
	bic	r4, r4, #0xff00 ; \
	/* carry */\
	and	scratch1, reg, #0x1 ; \
	orr	r4, r4, scratch1, lsl #8      ; \
	/* other flags */\
	ldr	scratch1, [r0, #44] ; \
	ldrb	scratch1, [scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8       ; \
	/* store result back in register */\
	bic	reg, reg, #0xff ; \
	orr	reg, reg, scratch2 ; \
	
	
#define BIT_TEST(reg, bit) \
	bic	r4, r4, #0xfe00 ; \
	orr	r4, r4, #0x1000 ; \
	tst	reg, bit 	; \
	orreq	r4, r4, #0x4400 ; \
	
	
#define BIT_TEST7(reg, bit) \
	bic	r4, r4, #0xfe00 ; \
	orr	r4, r4, #0x1000 ; \
	tst	reg, bit 	; \
	orreq	r4, r4, #0x4400 ; \
	orrne	r4, r4, #0x8000 ; \
	
#define ADD8(reg1, scratch1, scratch2) \
	/* Copy bits 3 and 7 of sources to bits {0,1} and {3,4}*/\
	and	scratch1, r4, #0x80 ; \
	mov	scratch2, scratch1, lsr #3 ; \
	and	scratch1, reg1, #0x80	   ; \
	orr	scratch2, scratch1, lsr #4 ; \
	and	scratch1, r4, #0x08	   ; \
	orr	scratch2, scratch1, lsr #2 ; \
	and	scratch1, reg1, #0x08	   ; \
	orr	scratch2, scratch1, lsr #3 ; \
	/* Clear all affected flags */\
	bic	r4, r4, #0xff00 ; \
	/* Actual addition */\
	add	r4, r4, reg1 	; \
	/* C is set by addition. Set N, 5, 3 */\
	and	scratch1, r4, #0xa8 ; \
	orr	r4, r4, scratch1, lsl #8 ; \
	/* set Z */\
	tst	r4, #0xff 	; \
	orreq	r4, r4, #0x4000 ; \
	/* Copy bits 3 and 7 of result to 2 and 5 */\
	and	scratch1, r4, #0x80 ; \
	orr	scratch2, scratch1, lsr #2 ; \
	and	scratch1, r4, #0x08	   ; \
	orr	scratch2, scratch1, lsr #1 ; \
	/* Set H and V */\
	ldr	scratch1, [r0, #48]	   ; \
	ldrb	scratch1, [scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8       ; \

	
#define ADC8(reg1, scratch1, scratch2) \
	/* Copy bits 3 and 7 of sources to bits {0,1} and {3,4}*/\
	and	scratch1, r4, #0x80 ; \
	mov	scratch2, scratch1, lsr #3 ; \
	and	scratch1, reg1, #0x80	   ; \
	orr	scratch2, scratch1, lsr #4 ; \
	and	scratch1, r4, #0x08	   ; \
	orr	scratch2, scratch1, lsr #2 ; \
	and	scratch1, reg1, #0x08	   ; \
	orr	scratch2, scratch1, lsr #3 ; \
	/* Test carry flag */\
	tst	r4, #0x100 	; \
	/* Clear all affected flags */\
	bic	r4, r4, #0xff00 ; \
	/* Add in carry if necessary */\
	add	r4, r4, reg1 	; \
	addne	r4, r4, #1	; \
	/* C is set by addition. Set N, 5, 3 */\
	and	scratch1, r4, #0xa8 ; \
	orr	r4, r4, scratch1, lsl #8 ; \
	/* set Z */\
	tst	r4, #0xff 	; \
	orreq	r4, r4, #0x4000 ; \
	/* Copy bits 3 and 7 of result to 2 and 5 */\
	and	scratch1, r4, #0x80 ; \
	orr	scratch2, scratch1, lsr #2 ; \
	and	scratch1, r4, #0x08	   ; \
	orr	scratch2, scratch1, lsr #1 ; \
	/* Set H and V */\
	ldr	scratch1, [r0, #48]	   ; \
	ldrb	scratch1, [scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8       ; \


#define SUB8(reg1, scratch1, scratch2) \
	/* Copy bits 3 and 7 of sources to bits {0,1} and {3,4}*/\
	and	scratch1, r4, #0x80 ; \
	mov	scratch2, scratch1, lsr #3 ; \
	and	scratch1, reg1, #0x80	   ; \
	orr	scratch2, scratch1, lsr #4 ; \
	and	scratch1, r4, #0x08	   ; \
	orr	scratch2, scratch1, lsr #2 ; \
	and	scratch1, reg1, #0x08	   ; \
	orr	scratch2, scratch1, lsr #3 ; \
	/* Clear all affected flags */\
	bic	r4, r4, #0xff00 ; \
	/* Perform subtraction */\
	orr	r4, r4, #0x100 	; \
	and	scratch1, reg1, #0xff ; \
	sub	r4, r4, scratch1      ; \
	eor	r4, r4, #0x100	      ; \
	/* C is set. Set N, 5, 3 */\
	and	scratch1, r4, #0xa8 ; \
	orr	r4, r4, scratch1, lsl #8 ; \
	/* set Z */\
	tst	r4, #0xff	; \
	orreq	r4, r4, #0x4000 ; \
	/* Set subtract flag */\
	orr	r4, r4, #0x200	; \
	/* Copy bits 3 and 7 of result to 2 and 5 */\
	and	scratch1, r4, #0x80 ; \
	orr	scratch2, scratch1, lsr #2 ; \
	and	scratch1, r4, #0x08	   ; \
	orr	scratch2, scratch1, lsr #1 ; \
	/* Set H and V */\
	ldr	scratch1, [r0, #52]	   ; \
	ldrb	scratch1, [scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8       ; \
	


#define CP8(reg1, scratch1, scratch2) \
	/* Copy bits 3 and 7 of sources to bits {0,1} and {3,4}*/\
	and	scratch1, r4, #0x80 ; \
	mov	scratch2, scratch1, lsr #3 ; \
	and	scratch1, reg1, #0x80	   ; \
	orr	scratch2, scratch1, lsr #4 ; \
	and	scratch1, r4, #0x08	   ; \
	orr	scratch2, scratch1, lsr #2 ; \
	and	scratch1, reg1, #0x08	   ; \
	orr	scratch2, scratch1, lsr #3 ; \
	/* Clear all affected flags */\
	bic	r4, r4, #0xff00 ; \
	sub	scratch1, r4, reg1 	; \
	/* Set C */\
	tst	scratch1, #0x100	 ; \
	orrne	r4, r4, #0x100		 ; \
	/* Set Z */\
	tst	scratch1, #0xff ; \
	orreq	r4, r4, #0x4000 ; \
	/* Set bits for flag lookup */\
	tst	scratch1, #0x80		  ; \
	orrne	scratch2, scratch2, #0x20 ; \
	tst	scratch1, #0x08		  ; \
	orrne	scratch2, scratch2, #0x4  ; \
	/* Set N, 5, 3 */\
	and	scratch1, scratch1, #0x80 ; \
	orr	r4, r4, scratch1, lsl #8 ; \
	and	scratch1, reg1, #0x28	 ; \
	orr	r4, r4, scratch1, lsl #8 ; \
	/* Set subtract flag */\
	orr	r4, r4, #0x200	; \
	/* Set H and V */\
	ldr	scratch1, [r0, #52]	   ; \
	ldrb	scratch1, [scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8       ; \
	

#define SBC8(reg1, scratch1, scratch2) \
	/* Copy bits 3 and 7 of sources to bits {0,1} and {3,4}*/\
	and	scratch1, r4, #0x80 ; \
	mov	scratch2, scratch1, lsr #3 ; \
	and	scratch1, reg1, #0x80	   ; \
	orr	scratch2, scratch1, lsr #4 ; \
	and	scratch1, r4, #0x08	   ; \
	orr	scratch2, scratch1, lsr #2 ; \
	and	scratch1, reg1, #0x08	   ; \
	orr	scratch2, scratch1, lsr #3 ; \
	/* Test carry flag */\
	tst	r4, #0x100 	; \
	/* Clear all affected flags */\
	bic	r4, r4, #0xff00 ; \
	/* Subtract carry if necessary */\
	and	scratch1, reg1, #0xff ; \
	orr	r4, r4, #0x100 	; \
	subne	r4, r4, #1	; \
	sub	r4, r4, scratch1 	; \
	eor	r4, r4, #0x100		; \
	/* C is set by subtraction. Set N, 5, 3 */\
	and	scratch1, r4, #0xa8 ; \
	orr	r4, r4, scratch1, lsl #8 ; \
	/* set Z */\
	tst	r4, #0xff	; \
	orreq	r4, r4, #0x4000 ; \
	/* Set subtract flag */\
	orr	r4, r4, #0x200	; \
	/* Copy bits 3 and 7 of result to 2 and 5 */\
	and	scratch1, r4, #0x80 ; \
	orr	scratch2, scratch1, lsr #2 ; \
	and	scratch1, r4, #0x08	   ; \
	orr	scratch2, scratch1, lsr #1 ; \
	/* Set H and V */\
	ldr	scratch1, [r0, #52]	   ; \
	ldrb	scratch1, [scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8       ; \
	

#define AND8(reg1, scratch1, scratch2) \
	and	scratch1, r4, reg1 ; \
	/* Clear flags and A, preserve R */\
	and	r4, r4, #0xff000000	; \
	/* Put result in A */\
	orr	r4, r4, scratch1	; \
	/* set the flags */\
	ldr	scratch2,[r0, #44]	; \
	ldrb	scratch2,[scratch2, scratch1] ; \
	orr	r4, r4, scratch2, lsl #8      ; \
	/* Set H flag */\
	orr	r4, r4, #0x1000 ; \


#define OR8(reg1, scratch1, scratch2) \
	orr	scratch1, r4, reg1 ; \
	and	scratch1, scratch1, #0xff ; \
	/* Clear flags and A, preserve R */\
	and	r4, r4, #0xff000000	; \
	/* Put result in A */\
	orr	r4, r4, scratch1	; \
	/* set the flags */\
	ldr	scratch2,[r0, #44] ; \
	ldrb	scratch2,[scratch2, scratch1] ; \
	orr	r4, r4, scratch2, lsl #8      ; \


#define XOR8(reg1, scratch1, scratch2) \
	eor	scratch1, r4, reg1 ; \
	and	scratch1, scratch1, #0xff ; \
	/* Clear flags and A, preserve R */\
	and	r4, r4, #0xff000000	; \
	/* Put result in A */\
	orr	r4, r4, scratch1	; \
	/* Set the flags */\
	ldr	scratch2,[r0, #44]	  ; \
	ldrb	scratch2,[scratch2, scratch1] ; \
	orr	r4, r4, scratch2, lsl #8      ; \



@ 16-bit arithmetic
#define INC16(reg) \
	add	reg, reg, #1 	; \
	bic	reg, reg, #65536 ; \
	
#define DEC16(reg) \
	sub	reg, reg, #1 	; \
	bic	reg, reg, #0xff000000 ; \
	bic	reg, reg, #0xff0000   ; \
	
#define ADD16(reg1, reg2, scratch1, scratch2) \
	/* copy bit 11 of sources to bits 1 and 0 */\
	and	scratch1, reg1, #0x800 ; \
	mov	scratch2, scratch1, lsr #10 ; \
	and	scratch1, reg2, #0x800	   ; \
	orr	scratch2, scratch1, lsr #11 ; \
	/* actual addition */\
	add	reg1, reg1, reg2 ; \
	/* copy bit 11 of result to bit 2*/\
	and	scratch1, reg1, #0x800 ; \
	orr	scratch2, scratch1, lsr #9 ; \
	/* need to set C, N and H - clear them first */\
	bic	r4, r4, #0x1300 ; \
	/* set C */\
	tst	reg1, #0x10000	; \
	orrne	r4, r4, #0x100 	; \
	/* set H from table */\
	ldr	scratch1, [r0, #48]  ; \
	ldrb	scratch1, [scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8       ; \
	bic	reg1, reg1, #0x10000 ; \

#define ADC16(reg1, scratch1, scratch2) \
	/* Copy bits 11 and 15 of sources to bits {0,1} and {3,4}*/\
	and	scratch1, r7, #0x8000 ; \
	mov	scratch2, scratch1, lsr #11 ; \
	and	scratch1, reg1, #0x8000	    ; \
	orr	scratch2, scratch1, lsr #12 ; \
	and	scratch1, r7, #0x0800	    ; \
	orr	scratch2, scratch1, lsr #10 ; \
	and	scratch1, reg1, #0x0800	    ; \
	orr	scratch2, scratch1, lsr #11 ; \
	/* Test carry flag */\
	tst	r4, #0x100 	; \
	/* Clear all affected flags */\
	bic	r4, r4, #0xff00 ; \
	/* Add in carry if necessary */\
	add	r7, r7, reg1 	; \
	addne	r7, r7, #1 	; \
	/* Set C*/\
	and	scratch1, r7, #0x10000 ; \
	orr	r4, r4, scratch1, lsr #8 ; \
	/* Set N, 5, 3 */\
	and	scratch1, r7, #0xa800 ; \
	orr	r4, r4, scratch1      ; \
	/* Set Z */\
	bic	r7, r7, #0x10000 ; \
	cmp	r7, #0		 ; \
	orreq	r4, r4, #0x4000	 ; \
	/* Copy bits 3 and 7 of result to 2 and 5 */\
	and	scratch1, r7, #0x8000 ; \
	orr	scratch2, scratch1, lsr #10 ; \
	and	scratch1, r7, #0x0800	    ; \
	orr	scratch2, scratch1, lsr #9  ; \
	/* Set H and V */\
	ldr	scratch1, [r0, #48] ; \
	ldrb	scratch1, [scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8       ; \
	
#define SBC16(reg1, scratch1, scratch2) \
	/* Copy bits 11 and 15 of sources to bits {0,1} and {3,4}*/\
	and	scratch1, r7, #0x8000 ; \
	mov	scratch2, scratch1, lsr #11 ; \
	and	scratch1, reg1, #0x8000	   ; \
	orr	scratch2, scratch1, lsr #12 ; \
	and	scratch1, r7, #0x0800	   ; \
	orr	scratch2, scratch1, lsr #10 ; \
	and	scratch1, reg1, #0x0800	   ; \
	orr	scratch2, scratch1, lsr #11 ; \
	/* Test carry flag */\
	tst	r4, #0x100 	; \
	/* Clear all affected flags */\
	bic	r4, r4, #0xff00 ; \
	/* Subtract carry if necessary */\
	sub	r7, r7, reg1 	; \
	subne	r7, r7, #1	; \
	/* Set C */\
	and	scratch1, r7, #0x10000 ; \
	orr	r4, r4, scratch1, lsr #8 ; \
	bic	r7, r7, #0xff0000		 ; \
	bic	r7, r7, #0xff000000		 ; \
	/* Set N, 5, 3 */\
	and	scratch1, r7, #0xa800 ; \
	orr	r4, r4, scratch1 ; \
	/* set Z */\
	cmp	r7, #0 		; \
	orreq	r4, r4, #0x4000 ; \
	/* Set subtract flag */\
	orr	r4, r4, #0x200	; \
	/* Copy bits 3 and 7 of result to 2 and 5 */\
	and	scratch1, r7, #0x8000 ; \
	orr	scratch2, scratch1, lsr #10 ; \
	and	scratch1, r7, #0x0800	   ; \
	orr	scratch2, scratch1, lsr #9 ; \
	/* Set H and V */\
	ldr	scratch1, [r0, #52]	   ; \
	ldrb	scratch1, [scratch1, scratch2] ; \
	orr	r4, r4, scratch1, lsl #8       ; \
	

	.text
	.code 32
	.align 2
	.global asmExecute

@============================================================
@
@ Main entry point into interpreter
@
@============================================================
	@ Register usage in here:
	@ r0 is the pointer to z80State (passed in as param)
	@    See mastersystem.c for structure defn and field offsets
	@ r1 is used to hold the insn_table pointer
	@ r2, r3, r12 are scratch
	@ r4 is AF - d31-24 = R, d23-16 = unused, d15-8 = F, d7-0 = A
	@ r5 is BC
	@ r6 is DE
	@ r7 is HL
	@ r8 is IX/IY, or scratch
	@ r9 is PC
	@ r10 is tstate counter
	@ r11 is readPages pointer
	@
	@ The interpreter will run until the T-states counter goes
	@ above zero, at which point it will return.
	@ A pending interrupt is checked for and executed once at the
	@ beginning.
asmExecute:
	push	{r4-r11,lr}  		@ preserve registers

	@ load Z80 registers etc. into ARM registers
	ldrh	r4,[r0,#12]		@ AF
	ldrh	r5,[r0,#14]		@ BC
	ldrh	r6,[r0,#16]		@ DE
	ldrh	r7,[r0,#18]		@ HL
	ldrh	r9,[r0,#22]		@ PC
	ldr	r10,[r0]		@ tstate counter
	ldr	r11,[r0,#4]		@ readPages pointer
	ldrb	r3,[r0,#37]		@ R register
	orr	r4, r4, r3, lsl #24	@ it goes in here

	ldr	r1,[r0,#40]		@ instruction table pointer

check_interrupt:
	@ check for NMI pending
	ldrb	r3, [r0, #86]
	cmp	r3, #0
	beq	no_nmi

	mov	r2, #0
	strb	r2, [r0, #86]		@ clear pending NMI flag

	@ stack PC
	ldrh	r12, [r0, #20]		@ SP
	DEC16(r12)
	mov	r9, r9, ror #8
	WRITE_BYTE(r9, r12, r3, r2)
	DEC16(r12)
	mov	r9, r9, ror #24
	WRITE_BYTE(r9, r12, r3, r2)
	strh	r12, [r0, #20]		@ update SP

	mov	r9, #0x66
	add	r10, r10, #12
	b	next_instruction
	
no_nmi:	
	
	@ check for interrupt pending
	ldrb	r3,[r0, #72]		@ int pending flag
	cmp	r3, #0
	beq	next_instruction	@ no interrupt, skip

	@ interrupt is pending - process it
	@ clear int pending flag
	mov	r2, #0
	strb	r2,[r0, #72]

	@ Gameboy CPU interrupts HALT even if interrupts are disabled
#ifdef GAMEBOY
	@ check if CPU was on a halt instruction, 'free' it if it was
	READ_BYTE(r3, r9, r3, r2)
	cmp	r3, #0x76
	addeq	r9, r9, #1
	bic	r9, r9, #0x10000
	mov	r2, #0
#endif
	
	@ check interrupts are enabled
	ldrb	r3,[r0, #38]		@ iff1
	cmp	r3, #0
	beq	next_instruction	@ disabled, skip

	@ disable them now
	strh	r2,[r0, #38]		@ iff1 & 2

#ifndef GAMEBOY
	@ check if CPU was on a halt instruction, 'free' it if it was
	READ_BYTE(r3, r9, r3, r2)
	cmp	r3, #0x76
	addeq	r9, r9, #1
	bic	r9, r9, #0x10000
#endif

	@ldrh	r3, [r0, #78]		@ paging
	@strh	r3, [r0, #96]		@ store
	
	@ stack PC
	ldrh	r12, [r0, #20]		@ SP
	DEC16(r12)
	mov	r9, r9, ror #8
	WRITE_BYTE(r9, r12, r3, r2)
	DEC16(r12)
	mov	r9, r9, ror #24
	WRITE_BYTE(r9, r12, r3, r2)
	strh	r12, [r0, #20]		@ update SP

	@ check interrupt mode
	mov	r9, #56			@ destination for im 0/1
	add	r10, r10, #13		@ time taken for im 0/1
	ldrb	r3, [r0, #73]		@ imode
	cmp	r3, #2
	bne	next_instruction

	@ im 2
	add	r10, r10, #6		@ extra time for im 2
	ldrb	r12, [r0, #36]		@ I register
	mov	r12, r12, lsl #8
	orr	r12, r12, #0xff
	READ_BYTE(r9, r12, r3, r2)	@ read PC from vector table
	INC16(r12)
	READ_BYTE(r3, r12, r3, r2)
	orr	r9, r9, r3, lsl #8
		
next_instruction:
	@ read a byte from PC into r3
	READ_BYTE(r3, r9, r3, r2)
	
	@ inc pc
	INC16(r9)

	@ inc R register here
	add	r4, r4, #16777216
	
	@ lookup instruction in table and branch to it
	ldr	pc,[r1, r3, lsl #2]

	.global	insn_table
insn_table:
	.word	insn00, insn01, insn02, insn03
	.word	insn04, insn05, insn06, insn07
	.word	insn08, insn09, insn0a, insn0b
	.word	insn0c, insn0d, insn0e, insn0f
	.word	insn10, insn11, insn12, insn13
	.word	insn14, insn15, insn16, insn17
	.word	insn18, insn19, insn1a, insn1b
	.word	insn1c, insn1d, insn1e, insn1f
	.word	insn20, insn21, insn22, insn23
	.word	insn24, insn25, insn26, insn27
	.word	insn28, insn29, insn2a, insn2b
	.word	insn2c, insn2d, insn2e, insn2f
	.word	insn30, insn31, insn32, insn33
	.word	insn34, insn35, insn36, insn37
	.word	insn38, insn39, insn3a, insn3b
	.word	insn3c, insn3d, insn3e, insn3f

	.word	insn40, insn41, insn42, insn43
	.word	insn44, insn45, insn46, insn47
	.word	insn48, insn49, insn4a, insn4b
	.word	insn4c, insn4d, insn4e, insn4f
	.word	insn50, insn51, insn52, insn53
	.word	insn54, insn55, insn56, insn57
	.word	insn58, insn59, insn5a, insn5b
	.word	insn5c, insn5d, insn5e, insn5f
	.word	insn60, insn61, insn62, insn63
	.word	insn64, insn65, insn66, insn67
	.word	insn68, insn69, insn6a, insn6b
	.word	insn6c, insn6d, insn6e, insn6f
	.word	insn70, insn71, insn72, insn73
	.word	insn74, insn75, insn76, insn77
	.word	insn78, insn79, insn7a, insn7b
	.word	insn7c, insn7d, insn7e, insn7f

	.word	insn80, insn81, insn82, insn83
	.word	insn84, insn85, insn86, insn87
	.word	insn88, insn89, insn8a, insn8b
	.word	insn8c, insn8d, insn8e, insn8f
	.word	insn90, insn91, insn92, insn93
	.word	insn94, insn95, insn96, insn97
	.word	insn98, insn99, insn9a, insn9b
	.word	insn9c, insn9d, insn9e, insn9f
	.word	insna0, insna1, insna2, insna3
	.word	insna4, insna5, insna6, insna7
	.word	insna8, insna9, insnaa, insnab
	.word	insnac, insnad, insnae, insnaf
	.word	insnb0, insnb1, insnb2, insnb3
	.word	insnb4, insnb5, insnb6, insnb7
	.word	insnb8, insnb9, insnba, insnbb
	.word	insnbc, insnbd, insnbe, insnbf

	.word	insnc0, insnc1, insnc2, insnc3
	.word	insnc4, insnc5, insnc6, insnc7
	.word	insnc8, insnc9, insnca, insncb
	.word	insncc, insncd, insnce, insncf
	.word	insnd0, insnd1, insnd2, insnd3
	.word	insnd4, insnd5, insnd6, insnd7
	.word	insnd8, insnd9, insnda, insndb
	.word	insndc, insndd, insnde, insndf
	.word	insne0, insne1, insne2, insne3
	.word	insne4, insne5, insne6, insne7
	.word	insne8, insne9, insnea, insneb
	.word	insnec, insned, insnee, insnef
	.word	insnf0, insnf1, insnf2, insnf3
	.word	insnf4, insnf5, insnf6, insnf7
	.word	insnf8, insnf9, insnfa, insnfb
	.word	insnfc, insnfd, insnfe, insnff

done_execute:
	@ we reached the end of a screen line
	ldr	r12, [r0, #80]		@ screenState pointer
	ldr	r3, [r12, #0]		@ scanline

	@
	@ Do things depending on the scanline. We should:
	@
	@  - render a screenline if it's visible
	@  - generate a sound sample
	@  - decrement the scanline counter if necessary, generate an
	@    interrupt and set the flag if it goes below 0
	@  - generate an interrupt and set the vsync flag if at end of
	@    frame
	@
	cmp	r3, #256
	bge	no_render
	cmp	r3, #64
	blt	no_render

	ldrb	r2, [r12, #22]		@ read VDP register 10
	streqb	r2, [r12, #33]		@ copy to scanline counter on line 64
	
	ldrb	r2, [r12, #33]		@ scanline counter
	subs	r2, r2, #1		@ count it down
	strb	r2, [r12, #33]
	bpl	no_scanline_int

	ldrb	r2, [r12, #22]		@ read VDP register 10
	strb	r2, [r12, #33]		@ copy to scanline counter

	mov	r2, #0xff
	strb	r2, [r12, #34]		@ set line int pending flag
	
no_scanline_int:
	@
	@ At this point:
	@  - r0 is still z80State pointer
	@  - r1 is still instruction table pointer
	@  - r2 is scratch
	@  - r3 is scanline (0-311)
	@  - r4-7 still contain CPU registers
	@  - r8 is scratch
	@  - r9 is still PC
	@  - r10 is still tstates
	@  - r11 is still page table pointer
	@  - r12 is screenState pointer
	@
	@ r0 we HAVE to preserve
	@ r1, r3, r11 can be easily reloaded afterwards
	@ r4-7, r9, r10 can be reloaded if we store them first
	@ r12 we're going to want
	@

	@ free up as many registers as possible
	strh	r4, [r0, #12]
	strh	r5, [r0, #14]
	strh	r6, [r0, #16]
	strh	r7, [r0, #18]
	strh	r9, [r0, #22]
	strh	r10, [r0, #0]

	@ render a scanline
#include "render.S"

	ldrh	r4, [r0, #12]
	ldrh	r5, [r0, #14]
	ldrh	r6, [r0, #16]
	ldrh	r7, [r0, #18]
	ldrh	r9, [r0, #22]
	ldrh	r10, [r0, #0]

	ldr	r1, [r0, #40]
	ldr	r3, [r12, #0]
	ldr	r11, [r0, #4]
	
no_render:
	@ generate sound sample
#include "soundgen.S"

	@ now check for line interrupts
	ldrb	r2, [r12, #12]		@ read VDP register 0
	tst	r2, #0x10		@ test scanline int enabled
	beq	no_assert_int

	ldrb	r2, [r12, #34]		@ line int pending
	strb	r2, [r0, #72]		@ interrupt pending

no_assert_int:	
	add	r3, r3, #1		@ increment scanline
	str	r3, [r12, #0]		@ and store again

	sub	r10, r10, #224		@ t-state counter
	
	cmp	r3, #312		@ loop back unless end of screen
	blt	check_interrupt		@ there might be a line interrupt

	@ end of frame stuff
	mov	r3, #0
	str	r3, [r12, #0]		@ reset scanline

	ldrb	r3, [r12, #13]		@ read VDP register 1
	tst	r3, #32			@ test if v-sync ints enabled
	mov	r3, #255
	strneb	r3, [r0, #72]		@ interrupt pending for next time
	
	ldrb	r3, [r12, #32]		@ read video status
	orr	r3, r3, #128		@ set vsync flag
	strb	r3, [r12, #32]

	@ldrb	r3, [r12, #22]		@ read VDP register 10
	@strb	r3, [r12, #33]		@ copy to scanline counter

done_frame:	
	@ store Z80 registers back into state struct
	strh	r4,[r0,#12]		@ AF
	strh	r5,[r0,#14]		@ BC
	strh	r6,[r0,#16]		@ DE
	strh	r7,[r0,#18]		@ HL
	strh	r9,[r0,#22]		@ PC
	str	r10,[r0]		@ tstate counter
	
	ldrb	r3,[r0,#37]		@ preserve bit 7 of R
	bic	r4, r4, #0x80000000
	and	r3, r3, #128
	orr	r3, r3, r4, lsr #24	@ and update the rest
	strb	r3,[r0,#37]
	
	pop	{r4-r11,lr}		@ restore caller's registers
	mov	r0,#9
	bx	lr

@============================================================
@
@ Instruction implementations
@
@============================================================
@ nop
insn00: END_INSN(#4)

@ ld bc,NN
insn01:
	@ fetch C
	@ read a byte from PC into C
	READ_BYTE(r5, r9, r3, r2)

	@ inc pc
	INC16(r9)

	@ fetch B into r3
	READ_BYTE(r3, r9, r3, r2)

	@ inc pc
	INC16(r9)

	@ combine B onto BC
	orr	r5, r5, r3, lsl #8

	END_INSN(#10)

	
@ ld (bc),a
insn02: WRITE_BYTE(r4, r5, r3, r2)
	END_INSN(#7)

	
@ inc bc
insn03:	INC16(r5)
	END_INSN(#6)

@ inc b
insn04: INC8_HI(r5, r2)
	END_INSN(#4)
	
@ dec b	
insn05: DEC8_HI(r5, r2)
	END_INSN(#4)

@ ld b, N
insn06: READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	and	r5, r5, #0xff
	orr	r5, r5, r3, lsl #8
	END_INSN(#7)

@ rlca	
insn07:	and	r3, r4, #0xff	@ isolate A contents
	bic	r4, r4, #0xff 	@ clear ready for rotate
	bic	r4, r4, #0x1300 @ clear H, N, C flags
	orr	r4, r4, r3, lsl #1
	orr	r4, r4, r3, lsr #7
	END_INSN(#4)

#ifndef GAMEBOY
@ ex af, af'
insn08: ldrh	r3,[r0, #28]    @ read af'
	strh	r4,[r0, #28]	@ store current AF in there
	and	r2, r4, #0xff000000  @ preserve R
	orr	r4, r3, r2
	END_INSN(#4)
#else
@ ld (NN),sp
insn08:	@ FIXME: implement
#endif

@ add hl, bc
insn09: ADD16(r7, r5, r3, r2)
	END_INSN(#11)
	
@ ld a,(bc)
insn0a: READ_BYTE(r3, r5, r3, r2)
	bic	r4, r4, #0xff
	orr	r4, r4, r3
	END_INSN(#7)

@ dec bc
insn0b: DEC16(r5)
	END_INSN(#6)

@ inc c
insn0c: INC8_LO(r5, r2)
	END_INSN(#4)

@ dec c	
insn0d: DEC8_LO(r5, r2)
	END_INSN(#4)

@ ld c,N
insn0e:	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	and	r5, r5, #0xff00
	orr	r5, r5, r3
	END_INSN(#7)

@ rrca
insn0f: and	r3, r4, #0xff
	mov	r2, r3, lsr #1
	mov	r3, r3, lsl #7
	and	r3, r3, #0x80
	bic	r4, r4, #0xff
	bic	r4, r4, #0x1300
	orr	r4, r4, r2
	orr	r4, r4, r3
	orr	r4, r4, r3, lsl #1
	END_INSN(#4)

#ifndef GAMEBOY
@ djnz DIS
insn10: READ_SIGNED_BYTE(r3, r9, r3, r2)
	INC16(r9)
	sub	r5, r5, #0x100
	bic	r5, r5, #0xff0000
	bic	r5, r5, #0xff000000
	tst	r5, #0xff00
	addne	r9, r9, r3
	addne	r10, r10, #5  @ extra cycles if branch taken
	bic	r9, #0xff0000
	bic	r9, #0xff000000
	END_INSN(#8)
#else
@ stop - meaning unknown
insn10:	END_INSN(#4)
#endif

@ ld de, NN
insn11:
	@ fetch E
	@ read a byte from PC into E
	READ_BYTE(r6, r9, r3, r2)

	@ inc pc
	INC16(r9)

	@ fetch D into r3
	READ_BYTE(r3, r9, r3, r2)

	@ inc pc
	INC16(r9)

	@ combine D onto DE
	orr	r6, r6, r3, lsl #8

	END_INSN(#10)

@ ld (de),a
insn12:	WRITE_BYTE(r4, r6, r3, r2)
	END_INSN(#7)

@ inc de
insn13: INC16(r6)
	END_INSN(#6)

@ inc d
insn14: INC8_HI(r6, r2)
	END_INSN(#4)

@ dec d
insn15: DEC8_HI(r6, r2)
	END_INSN(#4)

@ ld d,N
insn16: READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	and	r6, r6, #0xff
	orr	r6, r6, r3, lsl #8
	END_INSN(#7)

@ rla
insn17: and	r3, r4, #0xff
	and	r2, r4, #0x100
	mov	r3, r3, lsl #1
	orr	r3, r3, r2, lsr #8
	bic	r4, #0xff
	bic	r4, #0x1300
	orr	r4, r4, r3
	END_INSN(#4)
	
@ jr DIS
insn18: READ_SIGNED_BYTE(r3, r9, r3, r2)
	INC16(r9)
	add	r9, r9, r3
	bic	r9, #0xff0000
	bic	r9, #0xff000000
	END_INSN(#12)

@ add hl,de
insn19: ADD16(r7, r6, r3, r2)
	END_INSN(#11)

@ ld a,(de)
insn1a:	READ_BYTE(r3, r6, r3, r2)
	bic	r4, r4, #0xff
	orr	r4, r4, r3
	END_INSN(#7)

@ dec de
insn1b: DEC16(r6)
	END_INSN(#6)

@ inc e
insn1c: INC8_LO(r6, r2)
	END_INSN(#4)

@ dec e
insn1d:	DEC8_LO(r6, r2)
	END_INSN(#4)

@ ld e,N
insn1e: READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	and	r6, r6, #0xff00
	orr	r6, r6, r3
	END_INSN(#7)

@ rra
insn1f: mov	r3, r4, lsr #1
	and	r2, r4, #0x1
	and	r3, r3, #0xff
	bic	r4, r4, #0xff
	bic	r4, r4, #0x1300
	orr	r4, r4, r3
	orr	r4, r4, r2, lsl #8
	END_INSN(#4)

@ jr nz,DIS
insn20: READ_SIGNED_BYTE(r3, r9, r3, r2)
	INC16(r9)
	tst	r4, #0x4000
	addeq	r9, r9, r3
	addeq	r10, r10, #5
	bic	r9, #0xff0000
	bic	r9, #0xff000000
	END_INSN(#7)


@ ld hl,NN
insn21:
	@ fetch L
	@ read a byte from PC into L
	READ_BYTE(r7, r9, r3, r2)

	@ inc pc
	INC16(r9)

	@ fetch H into r3
	READ_BYTE(r3, r9, r3, r2)

	@ inc pc
	INC16(r9)

	@ combine H onto HL
	orr	r7, r7, r3, lsl #8

	END_INSN(#10)

#ifndef GAMEBOY
@ ld (NN),hl
insn22:	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	orr	r12, r12, r3, lsl #8
	WRITE_BYTE(r7, r12, r3, r2)
	INC16(r12)
	mov	r7, r7, ror #8
	WRITE_BYTE(r7, r12, r3, r2)
	mov	r7, r7, ror #24
	END_INSN(#16)
#else
@ ld (hli),a
insn22:	@ FIXME: implement
#endif

@ inc hl
insn23:	INC16(r7)
	END_INSN(#6)

@ inc h
insn24:	INC8_HI(r7, r2)
	END_INSN(#4)

@ dec h
insn25:	DEC8_HI(r7, r2)
	END_INSN(#4)

@ ld h,N
insn26:	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	and	r7, r7, #0xff
	orr	r7, r7, r3, lsl #8
	END_INSN(#7)

@ daa
insn27: @ work out correction factor in r12
	mov	r12, #0
	@ work out upper 4 bits
	tst	r4, #0x100
	orrne	r12, r12, #0x60
	and	r3, r4, #0xff		@ get just A
	cmp	r3, #0x99
	orrgt	r12, r12, #0x60
	orrgt	r4, r4, #0x100
	@ now work out lower 4 bits
	tst	r4, #0x1000
	orrne	r12, r12, #0x06
	and	r3, r4, #0x0f
	cmp	r3, #0x09
	orrgt	r12, r12, #0x06
	@ work out whether to add or subtract
	and	r3, r4, #0xff		@ get just A
	tst	r4, #0x200		@ test N flag
	subne	r3, r3, r12
	addeq	r3, r3, r12
	and	r3, r3, #0xff
	@ work out half-carry flag - bit 4 of r2
	eor	r2, r3, r4
	@ carry and N are already set
	bic	r4, r4, #0xfc00
	bic	r4, r4, #0xff
	@ put result in A
	orr	r4, r4, r3
	@ put half-carry flag in
	and	r2, r2, #0x10
	orr	r4, r4, r2, lsl #8
	@ S, Z, P, 5, 3 should be set as for logical operation - look up
	ldr	r2, [r0, #44]
	ldrb	r2, [r2, r3]
	orr	r4, r4, r2, lsl #8
	END_INSN(#4)

@ jr z,DIS
insn28: READ_SIGNED_BYTE(r3, r9, r3, r2)
	INC16(r9)
	tst	r4, #0x4000
	addne	r9, r9, r3
	addne	r10, r10, #5
	bic	r9, #0xff0000
	bic	r9, #0xff000000
	END_INSN(#7)


@ add hl,hl
insn29:	ADD16(r7, r7, r3, r2)
	END_INSN(#11)

#ifndef GAMEBOY
@ ld hl,(NN)
insn2a: READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	orr	r12, r12, r3, lsl #8
	READ_BYTE(r7, r12, r3, r2)
	INC16(r12)
	READ_BYTE(r3, r12, r3, r2)
	orr	r7, r7, r3, lsl #8
	END_INSN(#16)
#else
@ ld a,(hli)
insn2a:	@ FIXME: implement
#endif

@ dec hl
insn2b:	DEC16(r7)
	END_INSN(#6)

@ inc l
insn2c:	INC8_LO(r7, r2)
	END_INSN(#4)

@ dec l
insn2d:	DEC8_LO(r7, r2)
	END_INSN(#4)

@ ld l,N
insn2e:	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	and	r7, r7, #0xff00
	orr	r7, r7, r3
	END_INSN(#7)

@ cpl
insn2f:	eor	r4, r4, #0xff
	orr	r4, r4, #0x1200
	END_INSN(#4)

@ jr nc,DIS
insn30: READ_SIGNED_BYTE(r3, r9, r3, r2)
	INC16(r9)
	tst	r4, #0x100
	addeq	r9, r9, r3
	addeq	r10, r10, #5
	bic	r9, #0xff0000
	bic	r9, #0xff000000
	END_INSN(#7)

@ ld sp,NN
insn31: READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	orr	r12, r12, r3, lsl #8
	strh	r12,[r0,#20]
	END_INSN(#10)

#ifndef GAMEBOY
@ ld (NN),a
insn32:
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	orr	r12, r12, r3, lsl #8
	WRITE_BYTE(r4, r12, r3, r2)
	END_INSN(#13)
#else
@ ld (hld),a
insn32:	@ FIXME: implement
#endif

@ inc sp
insn33: ldrh	r12,[r0,#20]
	INC16(r12)
	strh	r12,[r0,#20]
	END_INSN(#6)

@ inc (hl)
insn34:	READ_BYTE(r12, r7, r3, r2)
	INC8_LO(r12, r2)
	WRITE_BYTE(r12, r7, r3, r2)
	END_INSN(#11)

@ dec(hl)
insn35: READ_BYTE(r12, r7, r3, r2)
	DEC8_LO(r12, r2)
	WRITE_BYTE(r12, r7, r3, r2)
	END_INSN(#11)

@ ld (hl),N
insn36:	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	WRITE_BYTE(r12, r7, r3, r2)
	END_INSN(#10)

@ scf
insn37: orr	r4, r4, #0x0100
	bic	r4, r4, #0x1200
	END_INSN(#4)

@ jr c,DIS
insn38: READ_SIGNED_BYTE(r3, r9, r3, r2)
	INC16(r9)
	tst	r4, #0x100
	addne	r9, r9, r3
	addne	r10, r10, #5
	bic	r9, #0xff0000
	bic	r9, #0xff000000
	END_INSN(#7)

@ add hl,sp
insn39: ldrh	r12,[r0, #20]
	ADD16(r7, r12, r3, r2)
	END_INSN(#11)

#ifndef GAMEBOY
@ ld a,(NN)
insn3a:	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	orr	r12, r12, r3, lsl #8
	READ_BYTE(r3, r12, r3, r2)
	bic	r4, r4, #0xff
	orr	r4, r4, r3
	END_INSN(#13)
#else
@ ld a,(hld)
insn3a:	@ FIXME: implement
#endif

@ dec sp
insn3b: ldrh	r12,[r0, #20]
	DEC16(r12)
	strh	r12,[r0, #20]
	END_INSN(#6)

@ inc a
insn3c: INC8_LO(r4, r2)
	END_INSN(#4)
	
@ dec a
insn3d:	DEC8_LO(r4, r2)
	END_INSN(#4)

@ ld a,N
insn3e:	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	bic	r4, r4, #0xff
	orr	r4, r4, r3
	END_INSN(#7)

@ ccf
insn3f: and	r3, r4, #0x100
	bic	r4, r4, #0x1200
	eor	r4, r4, #0x0100
	orr	r4, r4, r3, lsl #4
	END_INSN(#4)


@
@ LD group
@
insn40: END_INSN(#4) @ ld b,b
insn41: and	r5, r5, #0xff
	orr	r5, r5, r5, lsl #8  @ ld b,c
	END_INSN(#4)
insn42: and	r5, r5, #0xff
	and	r3, r6, #0xff00
	orr	r5, r5, r3          @ ld b,d
	END_INSN(#4)
insn43: and	r5, r5, #0xff
	and	r3, r6, #0xff
	orr	r5, r5, r3, lsl #8  @ ld b,e
	END_INSN(#4)
insn44: and	r5, r5, #0xff
	and	r3, r7, #0xff00
	orr	r5, r5, r3	@ ld b,h
	END_INSN(#4)
insn45: and	r5, r5, #0xff
	and	r3, r7, #0xff
	orr	r5, r5, r3, lsl #8 @ ld b,l
	END_INSN(#4)
insn46: READ_BYTE(r3, r7, r3, r2)
	and	r5, r5, #0xff
	orr	r5, r5, r3, lsl #8 @ ld b,(hl)
	END_INSN(#7)
insn47: and	r5, r5, #0xff
	and	r3, r4, #0xff
	orr	r5, r5, r3, lsl #8 @ ld b,a
	END_INSN(#4)
	
insn48: and	r5, r5, #0xff00
	orr	r5, r5, r5, lsr #8  @ ld c,b
	END_INSN(#4)
insn49: END_INSN(#4)     @ ld c,c
insn4a: and	r5, r5, #0xff00
	orr	r5, r5, r6, lsr #8  @ ld c,d
	END_INSN(#4)
insn4b: and	r3, r6, #0xff
	and	r5, r5, #0xff00
	orr	r5, r5, r3          @ ld c,e
	END_INSN(#4)
insn4c: and	r5, r5, #0xff00
	orr	r5, r5, r7, lsr #8  @ ld c,h
	END_INSN(#4)
insn4d: and	r5, r5, #0xff00
	and	r3, r7, #0xff
	orr	r5, r5, r3	@ ld c,l
	END_INSN(#4)
insn4e: and	r5, r5, #0xff00
	READ_BYTE(r3, r7, r3, r2)
	orr	r5, r5, r3	@ ld c,(hl)
	END_INSN(#7)
insn4f: and	r5, r5, #0xff00
	and	r3, r4, #0xff
	orr	r5, r5, r3	@ ld c,a
	END_INSN(#4)

insn50: and	r6, r6, #0xff
	and	r3, r5, #0xff00
	orr	r6, r6, r3	@ ld d,b
	END_INSN(#4)
insn51: and	r6, r6, #0xff
	and	r3, r5, #0xff
	orr	r6, r6, r3, lsl #8 @ ld d,c
insn52: END_INSN(#4)		@ ld d,d
insn53: and	r6, r6, #0xff
	orr	r6, r6, r6, lsl #8 @ ld d,e
	END_INSN(#4)
insn54: and	r6, r6, #0xff
	and	r3, r7, #0xff00
	orr	r6, r6, r3	@ ld d,h
	END_INSN(#4)
insn55: and	r6, r6, #0xff
	and	r3, r7, #0xff
	orr	r6, r6, r3, lsl #8 @ ld d,l
	END_INSN(#4)
insn56: READ_BYTE(r3, r7, r3, r2)
	and	r6, r6, #0xff
	orr	r6, r6, r3, lsl #8 @ ld d,(hl)
	END_INSN(#7)
insn57: and	r6, r6, #0xff
	and	r3, r4, #0xff
	orr	r6, r6, r3, lsl #8 @ ld d,a
	END_INSN(#4)

insn58: and	r6, r6, #0xff00
	orr	r6, r6, r5, lsr #8 	@ ld e,b
	END_INSN(#4)
insn59: and	r6, r6, #0xff00
	and	r3, r5, #0xff
	orr	r6, r6, r3		@ ld e,c
	END_INSN(#4)
insn5a: and	r6, r6, #0xff00
	orr	r6, r6, r6, lsr #8	@ ld e,d
	END_INSN(#4)
insn5b: END_INSN(#4)			@ ld e,e
insn5c: and	r6, r6, #0xff00
	orr	r6, r6, r7, lsr #8	@ ld e,h
	END_INSN(#4)
insn5d: and	r6, r6, #0xff00
	and	r3, r7, #0xff
	orr	r6, r6, r3		@ ld e,l
	END_INSN(#4)
insn5e: and	r6, r6, #0xff00
	READ_BYTE(r3, r7, r3, r2)
	orr	r6, r6, r3		@ ld e,(hl)
	END_INSN(#7)
insn5f: and	r6, r6, #0xff00
	and	r3, r4, #0xff
	orr	r6, r6, r3		@ ld e,a
	END_INSN(#4)

insn60: and	r7, r7, #0xff
	and	r3, r5, #0xff00
	orr	r7, r7, r3		@ ld h,b
	END_INSN(#4)
insn61: and	r7, r7, #0xff
	and	r3, r5, #0xff
	orr	r7, r7, r3, lsl #8	@ ld h,c
	END_INSN(#4)
insn62:	and	r7, r7, #0xff
	and	r3, r6, #0xff00
	orr	r7, r7, r3		@ ld h,d
	END_INSN(#4)
insn63: and	r7, r7, #0xff
	and	r3, r6, #0xff
	orr	r7, r7, r3, lsl #8	@ ld h,e
	END_INSN(#4)
insn64: END_INSN(#4)			@ ld h,h
insn65: and	r7, r7, #0xff
	orr	r7, r7, r7, lsl #8	@ ld h,l
	END_INSN(#4)
insn66: READ_BYTE(r3, r7, r3, r2)
	and	r7, r7, #0xff
	orr	r7, r7, r3, lsl #8	@ ld h,(hl)
	END_INSN(#7)
insn67:	and	r7, r7, #0xff
	and	r3, r4, #0xff
	orr	r7, r7, r3, lsl #8	@ ld h,a
	END_INSN(#4)

insn68:	and	r7, r7, #0xff00
	orr	r7, r7, r5, lsr #8	@ ld l,b
	END_INSN(#4)
insn69:	and	r7, r7, #0xff00
	and	r3, r5, #0xff
	orr	r7, r7, r3		@ ld l,c
	END_INSN(#4)
insn6a: and	r7, r7, #0xff00
	orr	r7, r7, r6, lsr #8	@ ld l,d
	END_INSN(#4)
insn6b: and	r7, r7, #0xff00
	and	r3, r6, #0xff
	orr	r7, r7, r3		@ ld l,e
	END_INSN(#4)
insn6c:	and	r7, r7, #0xff00
	orr	r7, r7, r7, lsr #8	@ ld l,h
	END_INSN(#4)
insn6d: END_INSN(#4)			@ ld l,l
insn6e: READ_BYTE(r3, r7, r3, r2)
	and	r7, r7, #0xff00
	orr	r7, r7, r3		@ ld l,(hl)
	END_INSN(#7)
insn6f:	and	r7, r7, #0xff00
	and	r3, r4, #0xff
	orr	r7, r7, r3		@ ld l,a
	END_INSN(#4)

insn70:	mov	r12, r5, lsr #8
	WRITE_BYTE(r12, r7, r3, r2)	@ ld (hl),b
	END_INSN(#7)
insn71:	WRITE_BYTE(r5, r7, r3, r2)	@ ld (hl),c
	END_INSN(#7)
insn72:	mov	r12, r6, lsr #8
	WRITE_BYTE(r12, r7, r3, r2)	@ ld (hl),d
	END_INSN(#7)
insn73:	WRITE_BYTE(r6, r7, r3, r2)	@ ld (hl),e
	END_INSN(#7)
insn74:	mov	r12, r7, lsr #8		@ ld (hl),h
	WRITE_BYTE(r12, r7, r3, r2)
	END_INSN(#7)
insn75:	WRITE_BYTE(r7, r7, r3, r2)	@ ld (hl),l
	END_INSN(#7)
insn76: @ HALT
	DEC16(r9)			@ point at HALT again
	mov	r10, #0			@ use up time
	b	done_execute
insn77:	WRITE_BYTE(r4, r7, r3, r2)	@ ld (hl),a
	END_INSN(#7)

insn78:	bic	r4, r4, #0xff
	orr	r4, r4, r5, lsr #8	@ ld a,b
	END_INSN(#4)
insn79:	bic	r4, r4, #0xff
	and	r3, r5, #0xff
	orr	r4, r4, r3		@ ld a,c
	END_INSN(#4)
insn7a:	bic	r4, r4, #0xff
	orr	r4, r4, r6, lsr #8	@ ld a,d
	END_INSN(#4)
insn7b:	bic	r4, r4, #0xff
	and	r3, r6, #0xff
	orr	r4, r4, r3		@ ld a,e
	END_INSN(#4)
insn7c:	bic	r4, r4, #0xff
	orr	r4, r4, r7, lsr #8	@ ld a,h
	END_INSN(#4)
insn7d:	bic	r4, r4, #0xff
	and	r3, r7, #0xff
	orr	r4, r4, r3		@ ld a,l
	END_INSN(#4)
insn7e:	READ_BYTE(r3, r7, r3, r2)
	bic	r4, r4, #0xff
	orr	r4, r4, r3		@ ld a,(hl)
	END_INSN(#7)
insn7f:	END_INSN(#4)			@ ld a,a

@
@ Arithmetic group
@
insn80:	mov	r12, r5, lsr #8		@ add a,b
	ADD8(r12, r2, r3)
	END_INSN(#4)
insn81: and	r12, r5, #0xff		@ add a,c
	ADD8(r12, r2, r3)
	END_INSN(#4)
insn82:	mov	r12, r6, lsr #8		@ add a,d
	ADD8(r12, r2, r3)
	END_INSN(#4)
insn83:	and	r12, r6, #0xff		@ add a,e
	ADD8(r12, r2, r3)
	END_INSN(#4)
insn84: mov	r12, r7, lsr #8		@ add a,h
	ADD8(r12, r2, r3)
	END_INSN(#4)
insn85:	and	r12, r7, #0xff		@ add a,l
	ADD8(r12, r2, r3)
	END_INSN(#4)
insn86:	READ_BYTE(r12, r7, r3, r2)	@ add a,(hl)
	ADD8(r12, r2, r3)
	END_INSN(#7)
insn87:	and	r12, r4, #0xff		@ add a,a
	ADD8(r12, r2, r3)
	END_INSN(#4)
	
insn88:	mov	r12, r5, lsr #8		@ adc a,b
	ADC8(r12, r2, r3)
	END_INSN(#4)
insn89: and	r12, r5, #0xff		@ adc a,c
	ADC8(r12, r2, r3)
	END_INSN(#4)
insn8a:	mov	r12, r6, lsr #8		@ adc a,d
	ADC8(r12, r2, r3)
	END_INSN(#4)
insn8b:	and	r12, r6, #0xff		@ adc a,e
	ADC8(r12, r2, r3)
	END_INSN(#4)
insn8c: mov	r12, r7, lsr #8		@ adc a,h
	ADC8(r12, r2, r3)
	END_INSN(#4)
insn8d:	and	r12, r7, #0xff		@ adc a,l
	ADC8(r12, r2, r3)
	END_INSN(#4)
insn8e:	READ_BYTE(r12, r7, r3, r2)	@ adc a,(hl)
	ADC8(r12, r2, r3)
	END_INSN(#7)
insn8f:	and	r12, r4, #0xff		@ adc a,a
	ADC8(r12, r2, r3)
	END_INSN(#4)


insn90:	mov	r12, r5, lsr #8		@ sub b
	SUB8(r12, r2, r3)
	END_INSN(#4)
insn91: and	r12, r5, #0xff		@ sub c
	SUB8(r12, r2, r3)
	END_INSN(#4)
insn92:	mov	r12, r6, lsr #8		@ sub d
	SUB8(r12, r2, r3)
	END_INSN(#4)
insn93:	and	r12, r6, #0xff		@ sub e
	SUB8(r12, r2, r3)
	END_INSN(#4)
insn94: mov	r12, r7, lsr #8		@ sub h
	SUB8(r12, r2, r3)
	END_INSN(#4)
insn95:	and	r12, r7, #0xff		@ sub l
	SUB8(r12, r2, r3)
	END_INSN(#4)
insn96:	READ_BYTE(r12, r7, r3, r2)	@ sub (hl)
	SUB8(r12, r2, r3)
	END_INSN(#7)
insn97:	and	r12, r4, #0xff		@ sub a
	SUB8(r12, r2, r3)
	END_INSN(#4)
	
insn98:	mov	r12, r5, lsr #8		@ sbc a,b
	SBC8(r12, r2, r3)
	END_INSN(#4)
insn99: and	r12, r5, #0xff		@ sbc a,c
	SBC8(r12, r2, r3)
	END_INSN(#4)
insn9a:	mov	r12, r6, lsr #8		@ sbc a,d
	SBC8(r12, r2, r3)
	END_INSN(#4)
insn9b:	and	r12, r6, #0xff		@ sbc a,e
	SBC8(r12, r2, r3)
	END_INSN(#4)
insn9c: mov	r12, r7, lsr #8		@ sbc a,h
	SBC8(r12, r2, r3)
	END_INSN(#4)
insn9d:	and	r12, r7, #0xff		@ sbc a,l
	SBC8(r12, r2, r3)
	END_INSN(#4)
insn9e:	READ_BYTE(r12, r7, r3, r2)	@ sbc a,(hl)
	SBC8(r12, r2, r3)
	END_INSN(#7)
insn9f:	and	r12, r4, #0xff		@ sbc a,a
	SBC8(r12, r2, r3)
	END_INSN(#4)


insna0:	mov	r12, r5, lsr #8		@ and b
	AND8(r12, r2, r3)
	END_INSN(#4)
insna1: and	r12, r5, #0xff		@ and c
	AND8(r12, r2, r3)
	END_INSN(#4)
insna2:	mov	r12, r6, lsr #8		@ and d
	AND8(r12, r2, r3)
	END_INSN(#4)
insna3:	and	r12, r6, #0xff		@ and e
	AND8(r12, r2, r3)
	END_INSN(#4)
insna4: mov	r12, r7, lsr #8		@ and h
	AND8(r12, r2, r3)
	END_INSN(#4)
insna5:	and	r12, r7, #0xff		@ and l
	AND8(r12, r2, r3)
	END_INSN(#4)
insna6:	READ_BYTE(r12, r7, r3, r2)	@ and (hl)
	AND8(r12, r2, r3)
	END_INSN(#7)
insna7:	and	r12, r4, #0xff		@ and a
	AND8(r12, r2, r3)
	END_INSN(#4)
	
insna8:	mov	r12, r5, lsr #8		@ xor b
	XOR8(r12, r2, r3)
	END_INSN(#4)
insna9: and	r12, r5, #0xff		@ xor c
	XOR8(r12, r2, r3)
	END_INSN(#4)
insnaa:	mov	r12, r6, lsr #8		@ xor d
	XOR8(r12, r2, r3)
	END_INSN(#4)
insnab:	and	r12, r6, #0xff		@ xor e
	XOR8(r12, r2, r3)
	END_INSN(#4)
insnac: mov	r12, r7, lsr #8		@ xor h
	XOR8(r12, r2, r3)
	END_INSN(#4)
insnad:	and	r12, r7, #0xff		@ xor l
	XOR8(r12, r2, r3)
	END_INSN(#4)
insnae:	READ_BYTE(r12, r7, r3, r2)	@ xor (hl)
	XOR8(r12, r2, r3)
	END_INSN(#7)
insnaf:	and	r12, r4, #0xff		@ xor a
	XOR8(r12, r2, r3)
	END_INSN(#4)


insnb0:	mov	r12, r5, lsr #8		@ or b
	OR8(r12, r2, r3)
	END_INSN(#4)
insnb1: and	r12, r5, #0xff		@ or c
	OR8(r12, r2, r3)
	END_INSN(#4)
insnb2:	mov	r12, r6, lsr #8		@ or d
	OR8(r12, r2, r3)
	END_INSN(#4)
insnb3:	and	r12, r6, #0xff		@ or e
	OR8(r12, r2, r3)
	END_INSN(#4)
insnb4: mov	r12, r7, lsr #8		@ or h
	OR8(r12, r2, r3)
	END_INSN(#4)
insnb5:	and	r12, r7, #0xff		@ or l
	OR8(r12, r2, r3)
	END_INSN(#4)
insnb6:	READ_BYTE(r12, r7, r3, r2)	@ or (hl)
	OR8(r12, r2, r3)
	END_INSN(#7)
insnb7:	and	r12, r4, #0xff		@ or a
	OR8(r12, r2, r3)
	END_INSN(#4)
	
insnb8:	mov	r12, r5, lsr #8		@ cp b
	CP8(r12, r2, r3)
	END_INSN(#4)
insnb9: and	r12, r5, #0xff		@ cp c
	CP8(r12, r2, r3)
	END_INSN(#4)
insnba:	mov	r12, r6, lsr #8		@ cp d
	CP8(r12, r2, r3)
	END_INSN(#4)
insnbb:	and	r12, r6, #0xff		@ cp e
	CP8(r12, r2, r3)
	END_INSN(#4)
insnbc: mov	r12, r7, lsr #8		@ cp h
	CP8(r12, r2, r3)
	END_INSN(#4)
insnbd:	and	r12, r7, #0xff		@ cp l
	CP8(r12, r2, r3)
	END_INSN(#4)
insnbe:	READ_BYTE(r12, r7, r3, r2)	@ cp (hl)
	CP8(r12, r2, r3)
	END_INSN(#7)
insnbf:	and	r12, r4, #0xff		@ cp a
	CP8(r12, r2, r3)
	END_INSN(#4)


insnc0:	@ ret nz
	add	r10,r10,#1
	tst	r4,#0x4000
	beq	insnc9
	END_INSN(#4)
insnc1:	@ pop bc
	ldrh	r12,[r0, #20]
	READ_BYTE(r5, r12, r3, r2)
	INC16(r12)
	READ_BYTE(r3, r12, r3, r2)
	INC16(r12)
	strh	r12,[r0, #20]
	orr	r5, r5, r3, lsl #8
	END_INSN(#10)
insnc2:	@ jp nz,NN
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	tst	r4, #0x4000
	orreq	r9, r12, r3, lsl #8
	END_INSN(#10)
insnc3: @ jp NN
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	READ_BYTE(r3, r9, r3, r2)
	orr	r9, r12, r3, lsl #8
	END_INSN(#10)
insnc4:	@ call nz,NN
	tst	r4, #0x4000
	beq	insncd
	INC16(r9)
	INC16(r9)
	END_INSN(#10)
insnc5:	@ push bc
	ldrh	r12,[r0, #20]
	DEC16(r12)
	mov	r5, r5, ror #8
	WRITE_BYTE(r5, r12, r3, r2)
	DEC16(r12)
	mov	r5, r5, ror #24
	WRITE_BYTE(r5, r12, r3, r2)
	strh	r12,[r0, #20]
	END_INSN(#11)
insnc6:	@ add a,N
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	ADD8(r12, r2, r3)
	END_INSN(#7)
insnc7:	@ rst 0
@	and	r2, r9, #0xff00
@	cmp	r2, #0xe800
@	bne	normalrst0
@	sub	r9, r9, #15
@	END_INSN(#4)
@normalrst0:	
	ldrh	r12,[r0, #20]
	DEC16(r12)
	mov	r9, r9, ror #8
	WRITE_BYTE(r9, r12, r3, r2)
	DEC16(r12)
	mov	r9, r9, ror #24
	WRITE_BYTE(r9, r12, r3, r2)
	strh	r12,[r0, #20]
	mov	r9, #0
	END_INSN(#11)
insnc8:	@ ret z
	add	r10,r10,#1
	tst	r4,#0x4000
	bne	insnc9
	END_INSN(#4)
insnc9:	@ ret

	@ldrh	r3, [r0, #96]
	@cmp	r9, #0xf2
	@streqh	r3, [r0, #78]
	@bleq	update_paging
	
	ldrh	r12,[r0, #20]
	READ_BYTE(r9, r12, r3, r2)
	INC16(r12)
	READ_BYTE(r3, r12, r3, r2)
	INC16(r12)
	strh	r12,[r0, #20]
	orr	r9, r9, r3, lsl #8
	END_INSN(#10)
insnca:	@ jp z,NN
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	tst	r4, #0x4000
	orrne	r9, r12, r3, lsl #8
	END_INSN(#10)
insncc:	@ call z,NN
	tst	r4, #0x4000
	bne	insncd
	INC16(r9)
	INC16(r9)
	END_INSN(#10)
insncd:	@ call NN
	@ read destination address
	READ_BYTE(r8, r9, r3, r2)
	INC16(r9)
	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	orr	r8, r8, r3, lsl #8
	@ stack old PC
	ldrh	r12,[r0, #20]
	mov	r9, r9, ror #8
	DEC16(r12)
	WRITE_BYTE(r9, r12, r3, r2)
	DEC16(r12)
	mov	r9, r9, ror #24
	WRITE_BYTE(r9, r12, r3, r2)
	strh	r12,[r0, #20]
	@ jump
	mov	r9, r8
	END_INSN(#17)
insnce:	@ adc a,N
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	ADC8(r12, r2, r3)
	END_INSN(#7)
insncf:	@ rst 8
	ldrh	r12,[r0, #20]
	DEC16(r12)
	mov	r9, r9, ror #8
	WRITE_BYTE(r9, r12, r3, r2)
	DEC16(r12)
	mov	r9, r9, ror #24
	WRITE_BYTE(r9, r12, r3, r2)
	strh	r12,[r0, #20]
	mov	r9, #8
	END_INSN(#11)

insnd0:	@ ret nc
	add	r10,r10,#1
	tst	r4,#0x100
	beq	insnc9
	END_INSN(#4)
insnd1:	@ pop de
	ldrh	r12,[r0, #20]
	READ_BYTE(r6, r12, r3, r2)
	INC16(r12)
	READ_BYTE(r3, r12, r3, r2)
	INC16(r12)
	strh	r12,[r0, #20]
	orr	r6, r6, r3, lsl #8
	END_INSN(#10)
insnd2:	@ jp nc,NN
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	tst	r4, #0x100
	orreq	r9, r12, r3, lsl #8
	END_INSN(#10)
#ifndef GAMEBOY
insnd3:	@ out (N),a
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	and	r3, r4, #0xff
	orr	r2, r12, r3, lsl #8
	bl	port_out
	END_INSN(#11)
#endif
insnd4:	@ call nc,NN
	tst	r4, #0x100
	beq	insncd
	INC16(r9)
	INC16(r9)
	END_INSN(#10)
insnd5:	@ push de
	ldrh	r12,[r0, #20]
	DEC16(r12)
	mov	r6, r6, ror #8
	WRITE_BYTE(r6, r12, r3, r2)
	DEC16(r12)
	mov	r6, r6, ror #24
	WRITE_BYTE(r6, r12, r3, r2)
	strh	r12,[r0, #20]
	END_INSN(#11)
insnd6:	@ sub N
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	SUB8(r12, r2, r3)
	END_INSN(#7)
insnd7:	@ rst 16
	ldrh	r12,[r0, #20]
	DEC16(r12)
	mov	r9, r9, ror #8
	WRITE_BYTE(r9, r12, r3, r2)
	DEC16(r12)
	mov	r9, r9, ror #24
	WRITE_BYTE(r9, r12, r3, r2)
	strh	r12,[r0, #20]
	mov	r9, #16
	END_INSN(#11)
insnd8:	@ ret c
	add	r10,r10,#1
	tst	r4,#0x100
	bne	insnc9
	END_INSN(#4)
#ifndef GAMEBOY
insnd9:	@ exx
	ldrh	r3,[r0, #30]	@ load bc'
	ldrh	r2,[r0, #32]	@ load de'
	ldrh	r12,[r0, #34]	@ load hl'
	strh	r5,[r0, #30]
	strh	r6,[r0, #32]
	strh	r7,[r0, #34]
	mov	r5,r3
	mov	r6,r2
	mov	r7,r12
	END_INSN(#4)
#else
insnd9:	@ reti
	@ FIXME: implement - enable interrupts and return
#endif
insnda:	@ jp c,NN
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	tst	r4, #0x100
	orrne	r9, r12, r3, lsl #8
	END_INSN(#10)
#ifndef GAMEBOY
insndb:	@ in a,(N)
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	and	r2, r4, #0xff
	orr	r2, r12, r2, lsl #8
	bl	port_in
	bic	r4, r4, #0xff
	orr	r4, r4, r2
	END_INSN(#11)
#endif
insndc:	@ call c,NN
	tst	r4, #0x100
	bne	insncd
	INC16(r9)
	INC16(r9)
	END_INSN(#10)
insnde:	@ sbc a,N
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	SBC8(r12, r2, r3)
	END_INSN(#7)
insndf:	@ rst 24
	ldrh	r12,[r0, #20]
	DEC16(r12)
	mov	r9, r9, ror #8
	WRITE_BYTE(r9, r12, r3, r2)
	DEC16(r12)
	mov	r9, r9, ror #24
	WRITE_BYTE(r9, r12, r3, r2)
	strh	r12,[r0, #20]
	mov	r9, #24
	END_INSN(#11)

#ifndef GAMEBOY
insne0:	@ ret po
	add	r10,r10,#1
	tst	r4,#0x400
	beq	insnc9
	END_INSN(#4)
#else
insne0:	@ ld (FFN),a
	@ FIXME: implement
#endif
insne1:	@ pop hl
	ldrh	r12,[r0, #20]
	READ_BYTE(r7, r12, r3, r2)
	INC16(r12)
	READ_BYTE(r3, r12, r3, r2)
	INC16(r12)
	strh	r12,[r0, #20]
	orr	r7, r7, r3, lsl #8
	END_INSN(#10)
#ifndef GAMEBOY
insne2:	@ jp po,NN
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	tst	r4, #0x400
	orreq	r9, r12, r3, lsl #8
	END_INSN(#10)
insne3:	@ ex (sp),hl
	ldrh	r12,[r0, #20]
	READ_BYTE(r8, r12, r3, r2)
	WRITE_BYTE(r7, r12, r3, r2)
	INC16(r12)
	READ_BYTE(r3, r12, r3, r2)
	orr	r8, r8, r3, lsl #8
	mov	r7, r7, ror #8
	WRITE_BYTE(r7, r12, r3, r2)
	mov	r7, r8
	END_INSN(#19)
insne4:	@ call po,NN
	tst	r4, #0x400
	beq	insncd
	INC16(r9)
	INC16(r9)
	END_INSN(#10)
#else
insne2:	@ ld (FFC),a
	@ FIXME: implement
#endif
insne5:	@ push hl

	ldrh	r12,[r0, #20]
	DEC16(r12)
	mov	r7, r7, ror #8
	WRITE_BYTE(r7, r12, r3, r2)
	DEC16(r12)
	mov	r7, r7, ror #24
	WRITE_BYTE(r7, r12, r3, r2)
	strh	r12,[r0, #20]
	END_INSN(#11)
insne6:	@ and N
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	AND8(r12, r2, r3)
	END_INSN(#7)
insne7:	@ rst 32
	ldrh	r12,[r0, #20]
	DEC16(r12)
	mov	r9, r9, ror #8
	WRITE_BYTE(r9, r12, r3, r2)
	DEC16(r12)
	mov	r9, r9, ror #24
	WRITE_BYTE(r9, r12, r3, r2)
	strh	r12,[r0, #20]
	mov	r9, #32
	END_INSN(#11)
#ifndef GAMEBOY
insne8:	@ ret pe
	add	r10,r10,#1
	tst	r4,#0x400
	bne	insnc9
	END_INSN(#4)
#else
insne8:	@ add sp,dis
	@ FIXME: implement
#endif
insne9:	@ jp (hl)
	mov	r9,r7
	END_INSN(#4)
#ifndef GAMEBOY
insnea:	@ jp pe,NN
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	tst	r4, #0x400
	orrne	r9, r12, r3, lsl #8
	END_INSN(#10)
insneb:	@ ex de,hl
	mov	r3,r7
	mov	r7,r6
	mov	r6,r3
	END_INSN(#4)
insnec:	@ call pe,NN
	tst	r4, #0x400
	bne	insncd
	INC16(r9)
	INC16(r9)
	END_INSN(#10)
#else
insnea:	@ ld (NN),a
	@ FIXME: implement
#endif
insnee:	@ xor N
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	XOR8(r12, r2, r3)
	END_INSN(#7)
insnef:	@ rst 40
	ldrh	r12,[r0, #20]
	DEC16(r12)
	mov	r9, r9, ror #8
	WRITE_BYTE(r9, r12, r3, r2)
	DEC16(r12)
	mov	r9, r9, ror #24
	WRITE_BYTE(r9, r12, r3, r2)
	strh	r12,[r0, #20]
	mov	r9, #40
	END_INSN(#11)

#ifndef GAMEBOY
insnf0:	@ ret p
	add	r10,r10,#1
	tst	r4,#0x8000
	beq	insnc9
	END_INSN(#4)
#else
insnf0:	@ ld a,(FFN)
	@ FIXME: implement
#endif
insnf1:	@ pop af
	ldrh	r12,[r0, #20]
	READ_BYTE(r8, r12, r3, r2)
	INC16(r12)
	READ_BYTE(r3, r12, r3, r2)
	INC16(r12)
	strh	r12,[r0, #20]
	and	r4, r4, #0xff000000
	orr	r4, r4, r3
	orr	r4, r4, r8, lsl #8
	END_INSN(#10)
#ifndef GAMEBOY
insnf2:	@ jp p,NN
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	tst	r4, #0x8000
	orreq	r9, r12, r3, lsl #8
	END_INSN(#10)
#endif
insnf3:	@ di
	mov	r2,#0
	strb	r2,[r0, #38]
	strb	r2,[r0, #39]
	END_INSN(#4)
#ifndef GAMEBOY
insnf4:	@ call p,NN
	tst	r4, #0x8000
	beq	insncd
	INC16(r9)
	INC16(r9)
	END_INSN(#10)
#endif
insnf5:	@ push af
	ldrh	r12,[r0, #20]
	DEC16(r12)
	WRITE_BYTE(r4, r12, r3, r2)
	DEC16(r12)
	mov	r4, r4, ror #8
	WRITE_BYTE(r4, r12, r3, r2)
	mov	r4, r4, ror #24
	strh	r12,[r0, #20]
	END_INSN(#11)
insnf6:	@ or N
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	OR8(r12, r2, r3)
	END_INSN(#7)
insnf7:	@ rst 48
	ldrh	r12,[r0, #20]
	DEC16(r12)
	mov	r9, r9, ror #8
	WRITE_BYTE(r9, r12, r3, r2)
	DEC16(r12)
	mov	r9, r9, ror #24
	WRITE_BYTE(r9, r12, r3, r2)
	strh	r12,[r0, #20]
	mov	r9, #48
	END_INSN(#11)
#ifndef GAMEBOY
insnf8:	@ ret m
	add	r10,r10,#1
	tst	r4,#0x8000
	bne	insnc9
	END_INSN(#4)
#else
insnf8:	@ ldhl sp,N
	@ FIXME: implement
#endif
insnf9:	@ ld sp,hl
	strh	r7,[r0, #20]
	END_INSN(#6)
#ifndef GAMEBOY
insnfa:	@ jp m,NN
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	READ_BYTE(r3, r9, r3, r2)
	INC16(r9)
	tst	r4, #0x8000
	orrne	r9, r12, r3, lsl #8
	END_INSN(#10)
#else
insnfa:	@ ld a,(NN)
	@ FIXME: implement
#endif
insnfb:	@ ei
	mov	r2,#0xff
	strb	r2,[r0, #38]
	strb	r2,[r0, #39]
	END_INSN(#4)
#ifndef GAMEBOY
insnfc:	@ call m,NN
	tst	r4, #0x8000
	bne	insncd
	INC16(r9)
	INC16(r9)
	END_INSN(#10)
#endif
insnfe:	@ cp N
	READ_BYTE(r12, r9, r3, r2)
	INC16(r9)
	CP8(r12, r2, r3)
	END_INSN(#7)
insnff:	@ rst 56
	ldrh	r12,[r0, #20]
	DEC16(r12)
	mov	r9, r9, ror #8
	WRITE_BYTE(r9, r12, r3, r2)
	DEC16(r12)
	mov	r9, r9, ror #24
	WRITE_BYTE(r9, r12, r3, r2)
	strh	r12,[r0, #20]
	mov	r9, #56
	END_INSN(#11)

	
@============================================================
@
@ Include the files of prefixed instructions
@
@============================================================
#include "z80cb.S"

#ifndef GAMEBOY
#include "z80ed.S"
#include "z80ix.S"
#include "z80ixcb.S"
#else
	@ These are no-ops on Gameboy CPU
insnd3:
insndb:	
insndd:
insne3:
insne4:
insneb:
insnec:	
insned:
insnf2:
insnf4:
insnfc:	
insnfd:
	END_INSN(#4)
#endif

@============================================================
@
@ Port I/O callbacks - machine specific
@
@============================================================
@
@ MS ports (most important):
@
@ 0x7E        - mirror of soundchip/scanline port 0x7F
@ 0x7F        - soundchip write, scanline read
@ 0xBD        - mirror of video status/address port 0xBF
@ 0xBE        - video data read/write
@ 0xBF        - video status read/address write
@ 0xC0 - 0xC1 - mirror of joypad ports 0xDC and 0xDD
@ 0xDC - 0xDD - joypad read
@
@ Can hopefully ignore the internationalisation one, the FM
@ ones and the unknown ones.
@

	@
	@ Call with 16-bit port number in r2
	@ Returns result in r2
	@ May clobber r3, r8, r12
	@
port_in:
	and	r2, r2, #0xff
	
	cmp	r2, #0xdc
	beq	port_in_jp1
	cmp	r2, #0xc0
	bne	port_in_not_jp1
port_in_jp1:
	ldrb	r2, [r0, #84]
	bx	lr
port_in_not_jp1:

	cmp	r2, #0xdd
	beq	port_in_jp2
	cmp	r2, #0xc1
	bne	port_in_not_jp2
port_in_jp2:
	ldrb	r2, [r0, #85]
	bx	lr
port_in_not_jp2:

	and	r3, r2, #0xfe
	cmp	r3, #0x7e
	bne	port_in_not_scanline
	ldr	r12, [r0, #80]       	@ get screenState
	ldr	r2, [r12, #0]		@ get scanline
	subs	r2, r2, #64		@ remove top border
	movmi	r2, #0xff
	cmp	r2, #192
	movge	r2, #0xff
	bx	lr
port_in_not_scanline:

	and	r3, r2, #0xfd
	cmp	r3, #0xbd
	bne	port_in_not_vid_status
	ldr	r12, [r0, #80]       	@ get screenState
	ldrb	r2, [r12, #32]		@ get video status byte
	and	r3, r2, #0x3f		@ clear 2 flags on read
	strb	r3, [r12, #32]
	mov	r3, #0
	strb	r3, [r12, #31]		@ clear the latch
	strb	r3, [r12, #34]		@ and line int pending flag
	bx	lr
port_in_not_vid_status:

	cmp	r2, #0xbe
	bne	port_in_not_vid_data
	ldr	r12, [r0, #80]       	@ get screenState
	mov	r3, #0
	strb	r3, [r12, #31]		@ clear the latch
	ldrh	r3, [r12, #28]		@ get the VRAM address
	tst	r3, #0x8000
	bne	port_in_palette
	cmp	r3, #0x4000
	bge	port_in_skip
	ldr	r2, [r12, #4]		@ VRAM base pointer
	ldrb	r2, [r2, r3]		@ read a byte
	add	r3, r3, #1		@ increment address
	bic	r3, r3, #0x8000
	strh	r3, [r12, #28]
	bx	lr
port_in_palette:
	bic	r8, r3, #0x8000
	cmp	r8, #0x20
	bge	port_in_skip
	ldr	r2, [r12, #8]		@ palette pointer
	ldrb	r2, [r2, r8]		@ read a byte
	add	r3, r3, #1
	bic	r3, r3, #0x20		@ increment address
	strh	r3, [r12, #28]
	bx	lr
port_in_not_vid_data:

port_in_skip:
	mov	r2, #0xff
	bx	lr

	
	@
	@ Call with 16-bit port number in r2 and value to output in r3
	@ May clobber r2, r3, r8, r12
	@
port_out:
	and	r12, r2, #0xfe
	cmp	r12, #0x7e
	bne	port_out_not_sound
	@
	@ Top bit of value is 0 for writing coarse frequency, 1 for
	@ everything else. Top 4 bits mean:
	@
	@ 1000 - channel A frequency
	@ 1001 - channel A volume
	@ 1010 - channel B frequency
	@ 1011 - channel B volume
	@ 1100 - channel C frequency
	@ 1101 - channel C volume
	@ 1110 - noise frequency etc.
	@ 1111 - noise volume
	@
	ldr	r12, [r0, #88]		@ soundState pointer
	tst	r3, #0x80
	beq	port_out_coarse_freq
	and	r2, r3, #0xf0
	cmp	r2, #0xe0
	beq	port_out_noise_freq
	tst	r2, #0x10
	beq	port_out_fine_freq
	@ setting amplitude
	and	r2, r2, #0x60
	mov	r2, r2, lsr #5		@ extract channel number
	and	r3, r3, #0xf
	strb	r3, [r12, r2]		@ store amplitude
	bx	lr
port_out_fine_freq:
	and	r2, r2, #0x60		@ extract channel number
	mov	r2, r2, lsr #5
	strb	r2, [r12, #11]		@ store selected channel number
	mov	r2, r2, lsl #1
	add	r2, r2, #4
	ldrh	r8, [r12, r2]		@ get current frequency
	and	r8, r8, #0x3f0
	and	r3, r3, #0xf
	orr	r8, r8, r3		@ construct new frequency
	strh	r8, [r12, r2]		@ store
	bx	lr
port_out_noise_freq:
	and	r3, r3, #7
	strb	r3, [r12, #10]		@ store noise config bits
	ands	r3, r3, #3		@ noise frequency
	mov	r2, #0x10
	streqh	r2, [r12, #34]
	mov	r2, #0x20
	cmp	r3, #1
	streqh	r2, [r12, #34]
	mov	r2, #0x40
	cmp	r3, #2
	streqh	r2, [r12, #34]
	ldrh	r2, [r12, #6]		@ channel B frequency
	cmp	r3, #3
	streqh	r2, [r12, #34]
	mov	r3, #0x8000
	strh	r3, [r12, #32]		@ initialise LFSR
	bx	lr
port_out_coarse_freq:
	ldrb	r2, [r12, #11]		@ selected channel
	mov	r2, r2, lsl #1
	add	r2, r2, #4
	ldrh	r8, [r12, r2]		@ read right channel's freq
	and	r8, r8, #15
	and	r3, r3, #63
	orr	r8, r8, r3, lsl #4	@ construct new freq
	strh	r8, [r12, r2]		@ store
	bx	lr
port_out_not_sound:

	and	r12, r2, #0xfd
	cmp	r12, #0xbd
	bne	port_out_not_video_addr

	ldr	r12, [r0, #80]		@ get screenState
	ldrb	r2, [r12, #31]		@ is something latched?
	cmp	r2, #0
	bne	port_out_video_addr2
	orr	r3, r3, #0xff00		@ store latch value and flag
	strh	r3, [r12, #30]
	bx	lr
	@ handle second write to video address reg
port_out_video_addr2:
	mov	r2, #0
	strb	r2, [r12, #31]		@ clear the latch

	and	r2, r3, #0xf0
	cmp	r2, #0xc0
	beq	port_out_palette_addr
	cmp	r2, #0x80
	beq	port_out_vdp_reg
	
	@tst	r3, #0x80
	@bne	port_out_not_vram
	ldrb	r2, [r12, #30]		@ get latched value
	and	r3, r3, #0x3f
	orr	r3, r2, r3, lsl #8	@ construct full VRAM address
	strh	r3, [r12, #28]		@ and store it
	bx	lr
port_out_not_vram:
	tst	r3, #0x40
	bne	port_out_palette_addr
port_out_vdp_reg:	
	and	r3, r3, #0xf		@ register number
	ldrb	r2, [r12, #30]		@ latched value
	add	r3, r3, #12
	strb	r2, [r12, r3]		@ store value

	@cmp	r3, #22
	@streqb	r2, [r12, #33]		@ update scanline counter???
	bx	lr
port_out_palette_addr:
	ldrb	r3, [r12, #30]		@ get address
	orr	r3, r3, #0x8000		@ indicate palette
	strh	r3, [r12, #28]
	bx	lr
port_out_not_video_addr:

	and	r12, r2, #0xff
	cmp	r12, #0xbe
	bne	port_out_not_video_data
	ldr	r12, [r0, #80]       	@ get screenState
	mov	r8, #0
	strb	r8, [r12, #31]		@ clear the latch
	ldrh	r8, [r12, #28]		@ get the VRAM address
	tst	r8, #0x8000
	bne	port_out_palette

	cmp	r8, #0x4000
	bge	port_out_skip
	
	ldr	r2, [r12, #4]		@ VRAM base pointer
	strb	r3, [r2, r8]		@ write a byte

	bic	r8, r8, #3		@ get word aligned address
	ldr	r3, [r2, r8]		@ read whole 32-bits
	ldr	r12, [r12, #44]		@ tileBitmaps
	add	r12, r12, r8, lsl #1	@ point to this tile

	@ convert the tile row to a more normal bitmap
	and	r2, r3, #0x80
	mov	r8, r2, lsr #7
	and	r2, r3, #0x8000
	orr	r8, r2, lsr #14
	and	r2, r3, #0x800000
	orr	r8, r2, lsr #21
	and	r2, r3, #0x80000000
	orr	r8, r2, lsr #28
	strb	r8, [r12], #1
	
	and	r2, r3, #0x40
	mov	r8, r2, lsr #6
	and	r2, r3, #0x4000
	orr	r8, r2, lsr #13
	and	r2, r3, #0x400000
	orr	r8, r2, lsr #20
	and	r2, r3, #0x40000000
	orr	r8, r2, lsr #27
	strb	r8, [r12], #1
	
	and	r2, r3, #0x20
	mov	r8, r2, lsr #5
	and	r2, r3, #0x2000
	orr	r8, r2, lsr #12
	and	r2, r3, #0x200000
	orr	r8, r2, lsr #19
	and	r2, r3, #0x20000000
	orr	r8, r2, lsr #26
	strb	r8, [r12], #1
	
	and	r2, r3, #0x10
	mov	r8, r2, lsr #4
	and	r2, r3, #0x1000
	orr	r8, r2, lsr #11
	and	r2, r3, #0x100000
	orr	r8, r2, lsr #18
	and	r2, r3, #0x10000000
	orr	r8, r2, lsr #25
	strb	r8, [r12], #1
	
	and	r2, r3, #0x8
	mov	r8, r2, lsr #3
	and	r2, r3, #0x800
	orr	r8, r2, lsr #10
	and	r2, r3, #0x80000
	orr	r8, r2, lsr #17
	and	r2, r3, #0x8000000
	orr	r8, r2, lsr #24
	strb	r8, [r12], #1
	
	and	r2, r3, #0x4
	mov	r8, r2, lsr #2
	and	r2, r3, #0x400
	orr	r8, r2, lsr #9
	and	r2, r3, #0x40000
	orr	r8, r2, lsr #16
	and	r2, r3, #0x4000000
	orr	r8, r2, lsr #23
	strb	r8, [r12], #1
	
	and	r2, r3, #0x2
	mov	r8, r2, lsr #1
	and	r2, r3, #0x200
	orr	r8, r2, lsr #8
	and	r2, r3, #0x20000
	orr	r8, r2, lsr #15
	and	r2, r3, #0x2000000
	orr	r8, r2, lsr #22
	strb	r8, [r12], #1
	
	and	r2, r3, #0x1
	mov	r8, r2
	and	r2, r3, #0x100
	orr	r8, r2, lsr #7
	and	r2, r3, #0x10000
	orr	r8, r2, lsr #14
	and	r2, r3, #0x1000000
	orr	r8, r2, lsr #21
	strb	r8, [r12]
	
	ldr	r12, [r0, #80]		@ get screenState back
	ldrh	r8, [r12, #28]		@ get the VRAM address back
	add	r8, r8, #1		@ increment address
	bic	r8, r8, #0x8000
	strh	r8, [r12, #28]
	bx	lr
port_out_palette:
	bic	r8, r8, #0x8000
	cmp	r8, #0x20
	bge	port_out_skip
	ldr	r2, [r12, #8]		@ palette pointer
	strb	r3, [r2, r8]		@ write a byte

	and	r3, r3, #0x3f
	ldr	r2, [r12, #36]		@ colour table
	ldr	r3, [r2, r3, lsl #2]	@ convert colour
	ldr	r2, [r12, #40]		@ converted palette pointer
	str	r3, [r2, r8, lsl #2]	@ store converted colour
	
	add	r8, r8, #1
	@bic	r8, r8, #0x20		@ increment address
	orr	r8, r8, #0x8000
	strh	r8, [r12, #28]		@ store new address
	bx	lr
port_out_not_video_data:
	cmp	r12, #0x3f
	bne	port_out_not_3f
	ldrb	r2, [r0, #85]		@ joypad port 2
	and	r2, r2, #0x3f
	and	r12, r3, #0x20
	orr	r2, r2, r12, lsl #1
	and	r12, r3, #0x80
	orr	r2, r2, r12
	strb	r3, [r0, #87]
	strb	r2, [r0, #85]
	bx	lr

port_out_not_3f:
	bx	lr
port_out_skip:
	bx	lr
	

	@ this gets called from WRITE_BYTE so shouldn't clobber anything
	@ FIXME: might need to clip ROM pages to valid ones, not sure yet
update_paging:
	push	{r5-r8}

	ldr	r5, [r11, #64]		@ ROM start pointer
	ldrb	r8, [r0, #75]		@ ROM mask

	@ do ROM bank 0 paging
	ldrb	r6, [r0, #77]		@ paging reg 1
	and	r6, r6, r8
	add	r7, r5, r6, lsl #14
	str	r7, [r11, #0]
	add	r7, r7, #0x2000
	str	r7, [r11, #4]

	@ do ROM bank 1 paging
	ldrb	r6, [r0, #78]		@ paging reg 2
	and	r6, r6, r8
	add	r7, r5, r6, lsl #14
	str	r7, [r11, #8]
	add	r7, r7, #0x2000
	str	r7, [r11, #12]

	@ do ROM bank 2/RAM paging
	ldrb	r6, [r0, #76]		@ paging reg 0
	tst	r6, #8
	bne	paging_ram
	ldrb	r6, [r0, #79]		@ paging reg 3
	and	r6, r6, r8
	add	r7, r5, r6, lsl #14
	str	r7, [r11, #16]
	add	r7, r7, #0x2000
	str	r7, [r11, #20]

	@ ensure ROM writes don't happen
	ldr	r7, [r11, #32]		@ write scratch page
	str	r7, [r11, #48]
	str	r7, [r11, #52]

	pop	{r5-r8}
	bx	lr


	@ cartridge RAM is paged in
paging_ram:
	ldr	r5, [r11, #72]		@ cartridge RAM
	and	r6, r6, #4		@ bit 2 (RAM select)
	add	r7, r5, r6, lsl #12
	str	r7, [r11, #16]
	str	r7, [r11, #48]		@ store in read and write banks
	add	r7, r7, #0x2000
	str	r7, [r11, #20]
	str	r7, [r11, #52]

	pop	{r5-r8}
	bx	lr
